import re
from copy import deepcopy
from dataclasses import dataclass
from datetime import date, datetime, time, tzinfo
from enum import Enum
from typing import Any, Dict, Union

from kayaku.backend import dumps, loads
from kayaku.backend.types import JObject, JWrapper
from kayaku.pretty import Prettifier
from kayaku.utils import from_dict, update, copying_field

update_input = """\
{
    "a": "b", // sigma
    // preserved
    "d": "e",
    //lost
    "f": "g",
    // updated
    "h": "i",
    "j": [
        /*nice*/
        "viva",
        {} /*test*/,
        "deleted"
    ],
    "p": [],
    "v": {"a": 5},
    "q": {/*comment*/"p": 1}
}
"""

update_output = """\
{
    "a": "b", // sigma
    // preserved
    "d": "e",
    //lost
    "f": "g",
    // updated
    "h": 4,
    "j": [
        /*nice*/
        "viva",
        /*test*/
        {}
    ],
    "p": [
        1,
        2,
        {"a": "b"}
    ],
    "v": {"a": 5},
    "q": {"a": 1},
    "k": "2022-01-01T08:00:00",
    "re": "(\\\\d+)"
}"""


def test_json_update():
    o_obj = loads(update_input)
    obj = deepcopy(o_obj)
    del obj["f"]
    obj["h"] = 4
    obj["j"].pop()
    obj["k"] = datetime(2022, 1, 1, 8, 0, 0)
    obj["re"] = re.compile(r"(\d+)")
    obj["p"].extend([1, 2, {"a": "b"}])
    obj["q"] = Sub(1)
    update(o_obj, obj)
    assert dumps(Prettifier().prettify(o_obj)) == update_output


class E(Enum):
    A = 1
    B = 2
    C = 3


@dataclass
class Obj:
    a: E
    b: date
    c: time
    d: datetime
    e: re.Pattern
    f: Union[list, bool]
    g: Union[int, None]
    h: Union[int, str]


@dataclass
class Sub:
    a: int


@dataclass
class Obj2:
    a: E
    b: date
    c: time
    d: datetime
    e: re.Pattern
    f: Union[list, bool]
    g: Union[int, None]
    h: Union[int, str]
    i: Dict[str, Union[None, Any]]


def test_extra_load():
    dt_now = datetime.now()

    obj = {
        "a": 1,
        "b": f"{dt_now.date().isoformat()}",
        "c": f"{dt_now.time().isoformat()}",
        "d": f"{dt_now.isoformat()}",
        "e": r"(?P<name>\d+)",
        "f": JWrapper(True),
        "g": JWrapper(None),
        "h": "abc",
        "i": {"a": "a", "b": None},
    }
    target_obj = Obj2(
        E.A,
        dt_now.date(),
        dt_now.time(),
        dt_now,
        re.compile(r"(?P<name>\d+)"),
        True,
        None,
        "abc",
        {"a": "a", "b": None},
    )
    o: Obj2 = from_dict(Obj2, obj)
    assert o == target_obj
    assert o.f is True
    assert o.g is None
    assert o.i["b"] is None


def test_update_with_dc():
    from kayaku.backend.types import Array, Integer, JObject, JString, JWrapper

    dt_now = datetime.now()
    target_obj = Obj(
        E.A,
        dt_now.date(),
        dt_now.time(),
        dt_now,
        re.compile(r"(?P<name>\d+)"),
        [E.A, True, re.compile(r"(?P<name>.*?)"), dt_now, Sub(1), [5]],
        None,
        "abc",
    )
    o = JObject()
    update(o, target_obj)
    for k, v in JObject(
        {
            "a": Integer(1),
            "b": JString(dt_now.date().isoformat()),
            "c": JString(dt_now.time().isoformat()),
            "d": JString(dt_now.isoformat()),
            "e": JString("(?P<name>\\d+)"),
            "f": Array(
                [
                    Integer(1),
                    JWrapper(True),
                    JString(r"(?P<name>.*?)"),
                    JString(dt_now.isoformat()),
                    JObject({"a": Integer(1)}),
                    Array([Integer(5)]),
                ]
            ),
            "g": JWrapper(None),
            "h": JString("abc"),
        }
    ).items():
        assert o[k] == v
        assert o[k].__class__ == v.__class__


def test_copying_field():
    @dataclass
    class DC:
        ls: list = copying_field([])

    assert DC() == DC([])



import shutil
from pathlib import Path

import pytest

base_pth = Path("./temp/storage/").resolve()
if base_pth.exists():
    shutil.rmtree(base_pth.as_posix())
base_pth.mkdir(parents=True, exist_ok=True)


def test_insert_spec():
    import kayaku.storage
    from kayaku.spec import PathSpec, SectionSpec, SourceSpec

    root = kayaku.storage._PrefixNode()
    kayaku.storage._root.set(root)
    empty = SectionSpec([], [])

    kayaku.storage.insert(
        SourceSpec(["a", "b", "c"], ["d", "e", "f"], empty),
        PathSpec([*base_pth.parts, "config"], []),
    )

    kayaku.storage.insert(
        SourceSpec(["a", "b", "c"], ["d", "e", "credential"], empty),
        PathSpec([*base_pth.parts, "credential"], []),
    )

    with pytest.raises(ValueError):
        kayaku.storage.insert(
            SourceSpec(["a", "b", "c"], ["d", "e", "credential"], empty),
            PathSpec([*base_pth.parts, "credential"], []),
        )


def test_lookup_spec():
    import kayaku.spec
    import kayaku.storage
    from kayaku.spec import PathFill, PathSpec, SectionSpec, SourceSpec

    empty = SectionSpec([], [])

    root = kayaku.storage._PrefixNode()
    kayaku.storage._root.set(root)
    path_sect = [PathFill.EXTEND]

    kayaku.storage.insert(
        SourceSpec(["a", "b", "c"], ["d", "e", "f"], empty),
        PathSpec([*base_pth.parts, "config"], path_sect),
    )

    kayaku.storage.insert(
        SourceSpec(
            ["a", "b", "c", "xxx"],
            [],
            empty,
        ),
        PathSpec([*base_pth.parts, "hmm"], path_sect),
    )

    kayaku.storage.insert(
        SourceSpec(["a", "b", "c"], ["d", "e", "credential"], empty),
        PathSpec([*base_pth.parts, "credential"], path_sect),
    )

    kayaku.storage.insert(
        SourceSpec(["a", "b", "c", "d"], ["e", "f"], empty),
        PathSpec([*base_pth.parts, "any"], path_sect),
    )

    kayaku.storage.insert(
        SourceSpec(["a", "b", "c", "d"], ["o", "p", "e", "f"], empty),
        PathSpec([*base_pth.parts, "whirl"], path_sect),
    )

    assert (p := root.lookup(["a", "b", "c", "xxx", "d", "e", "f"])) and p[0] == (
        SourceSpec(["a", "b", "c", "xxx"], [], empty),
        PathSpec([*base_pth.parts, "hmm"], path_sect),
    )

    assert (p := root.lookup(["a", "b", "c", "d", "e", "credential"])) and p[0] == (
        SourceSpec(["a", "b", "c"], ["d", "e", "credential"], empty),
        PathSpec([*base_pth.parts, "credential"], path_sect),
    )

    assert (p := root.lookup(["a", "b", "c", "d", "p", "xxx", "e", "f"])) and p[0] == (
        SourceSpec(["a", "b", "c", "d"], ["e", "f"], empty),
        PathSpec([*base_pth.parts, "any"], path_sect),
    )

    assert (p := root.lookup(["a", "b", "c", "d", "o", "p", "e", "f"])) and p[0] == (
        SourceSpec(["a", "b", "c", "d"], ["o", "p", "e", "f"], empty),
        PathSpec([*base_pth.parts, "whirl"], path_sect),
    )

    assert root.lookup(["a", "b", "c", "d", "rand"]) is None

    assert root.lookup(["a", "b", "c", "d", "xxx", "f"]) is None


def test_spec_lookup_fmt_err():
    import kayaku.storage
    from kayaku.spec import FormattedPath, parse_path, parse_source

    root = kayaku.storage._PrefixNode()
    kayaku.storage._root.set(root)
    kayaku.storage.insert(
        parse_source("a.b.c.{**}"),
        parse_path(base_pth.as_posix() + "/a/b/c::{}"),
    )

    kayaku.storage.insert(
        parse_source("a.b.{**}"),
        parse_path(base_pth.as_posix() + "/d/e/f::{**}"),
    )

    assert (p := root.lookup(["a", "b", "c", "d", "e"])) and p[1] == FormattedPath(
        Path(base_pth, "d/e/f"), ["c", "d", "e"]
    )


def test_spec_lookup_wrapped():
    import kayaku.storage
    from kayaku.spec import FormattedPath, parse_path, parse_source

    root = kayaku.storage._PrefixNode()
    kayaku.storage._root.set(root)
    kayaku.storage.insert(
        parse_source("a.b.c.{**}"),
        parse_path(base_pth.as_posix() + "/a/b/c::{}"),
    )
    with pytest.raises(ValueError):
        kayaku.storage.lookup(["a", "b", "c", "d", "e"])

    kayaku.storage.insert(
        parse_source("a.b.{**}"),
        parse_path(base_pth.as_posix() + "/d/e/f.jsonc::{**}"),
    )

    assert kayaku.storage.lookup(["a", "b", "c", "d", "e"]) == FormattedPath(
        Path(base_pth, "d/e/f.jsonc"), ["c", "d", "e"]
    )



from __future__ import annotations

from dataclasses import dataclass, field
from dataclasses import fields as get_fields
from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple, Type

from .backend import dumps, loads
from .backend.types import JObject
from .format import format_with_model
from .pretty import Prettifier
from .schema_gen import ConfigModel, SchemaGenerator, write_schema_ref
from .spec import FormattedPath, parse_path, parse_source
from .storage import insert, lookup
from .utils import from_dict, update

MountType = Tuple[str, ...]
DomainType = Tuple[str, ...]


@dataclass
class _FileStore:
    generator: SchemaGenerator
    schemas: dict = field(default_factory=dict)
    mount: Dict[MountType, List[DomainType]] = field(default_factory=dict)
    field_mount_record: Set[MountType] = field(default_factory=set)

    def get_schema(self) -> dict:
        return {
            "$schema": "https://json-schema.org/draft/2020-12/schema",
            **self.schemas,
            "$defs": self.generator.defs,
        }


@dataclass
class _ModelStore:
    cls: Type[ConfigModel]
    location: FormattedPath
    instance: Optional[ConfigModel]


@dataclass
class _GlobalStore:
    prettifier: Prettifier
    generator_cls: Type[SchemaGenerator]
    files: Dict[Path, _FileStore] = field(default_factory=dict)
    models: Dict[DomainType, _ModelStore] = field(default_factory=dict)
    cls_domains: Dict[Type[ConfigModel], DomainType] = field(default_factory=dict)


_store: _GlobalStore


def insert_domain(domain: DomainType, cls: Type[ConfigModel]) -> None:
    _store.cls_domains[cls] = domain
    fmt_path: FormattedPath = lookup(list(domain))
    path = fmt_path.path
    mount_dest = tuple(fmt_path.mount_dest)
    file_store = _store.files.setdefault(path, _FileStore(_store.generator_cls(None)))
    for field in get_fields(cls):
        sub_dest: MountType = mount_dest + (field.name,)
        if sub_dest in file_store.field_mount_record:
            raise NameError(f"{path.as_posix()}::{'.'.join(sub_dest)} is occupied!")
        file_store.field_mount_record.add(sub_dest)
    file_store.mount.setdefault(mount_dest, []).append(domain)
    _store.models[domain] = _ModelStore(cls, fmt_path, None)
    file_store.generator.get_dc_schema(cls)
    write_schema_ref(
        file_store.schemas, mount_dest, file_store.generator.retrieve_name(cls)
    )


def initialize(
    specs: Dict[str, str],
    prettifier: Optional[Prettifier] = None,
    schema_generator_cls: Type[SchemaGenerator] = SchemaGenerator,
) -> None:
    """初始化 Kayaku.

    本函数需要最先被调用.

    Args:
        specs (Dict[str, str]): `domain 样式 -> 路径样式` 的映射.
        prettifier (Prettifier, optional): 格式化器.
        schema_generator_cls (Type[SchemaGenerator], optional): JSON Schema 生成器的类.

    Example:

        ```py
        from kayaku import bootstrap, config, initialize

        initialize({"{**}.connection": "./config/connection.jsonc::{**}})

        @config("my_mod.config.connection")
        class Connection:
            account: int | None = None
            "Account"
            password: str | None = None
            "password"

        bootstrap()
        ```

        以上代码将会将 `Connection` 类的数据存储在 `./config/connection.jsonc` 文件的 `["my_mod"]["config"]` 里.
    """
    global _store
    _store = _GlobalStore(prettifier or Prettifier(), schema_generator_cls)
    exceptions: list[Exception] = []
    for src, path in specs.items():
        try:
            src_spec = parse_source(src)
            path_spec = parse_path(path)
            insert(src_spec, path_spec)
        except Exception as e:
            exceptions.append(e)
    if exceptions:
        raise ValueError(
            f"{len(exceptions)} occurred during spec initialization.", exceptions
        )


def bootstrap() -> None:
    """预处理所有 `ConfigModel` 并写入默认值和 JSON Schema.

    建议在加载完外部模块后调用.
    """

    exceptions = []

    for path, store in _store.files.items():
        document = loads(path.read_text("utf-8") or "{}")
        path.with_suffix(".schema.json").write_text(
            dumps(store.get_schema()), encoding="utf-8"
        )
        for mount_dest, domains in store.mount.items():
            container = document
            for sect in mount_dest:
                container = container.setdefault(sect, JObject())
            for domain in domains:
                model_store = _store.models[domain]
                if model_store.instance is None:
                    try:
                        model_store.instance = from_dict(model_store.cls, container)
                    except Exception as e:
                        exceptions.append((path, mount_dest, model_store.cls, e))
                format_with_model(container, model_store.cls)
        document.pop("$schema", None)
        document["$schema"] = path.with_suffix(".schema.json").as_uri()
        path.write_text(
            dumps(_store.prettifier.prettify(document), endline=True), "utf-8"
        )
    if exceptions:
        raise ValueError(exceptions)


def save_all() -> None:
    """保存所有容器，并格式化。

    可以在退出前调用本函数。
    """

    for path, store in _store.files.items():
        document = loads(path.read_text("utf-8") or "{}")
        path.with_suffix(".schema.json").write_text(
            dumps(store.get_schema()), encoding="utf-8"
        )
        for mount_dest, domains in store.mount.items():
            container = document
            for sect in mount_dest:
                container = container.setdefault(sect, JObject())
            for domain in domains:
                model_store = _store.models[domain]
                if model_store.instance is not None:
                    update(container, model_store.instance)
                format_with_model(container, model_store.cls)
        document.pop("$schema", None)
        document["$schema"] = path.with_suffix(".schema.json").as_uri()
        path.write_text(
            dumps(_store.prettifier.prettify(document), endline=True), "utf-8"
        )



from __future__ import annotations

import inspect
from dataclasses import MISSING, Field, asdict, is_dataclass
from typing import Union, cast

from .backend.types import Array, BlockStyleComment, JObject, JString, JType, convert
from .schema_gen import ConfigModel


def remove_generated_comment(obj: JType):
    obj.json_before = [wsc for wsc in obj.json_before if "@type" not in wsc]
    obj.json_after = [wsc for wsc in obj.json_after if "@type" not in wsc]
    if isinstance(obj, (JObject, Array)):
        obj.json_container_tail = [
            wsc for wsc in obj.json_container_tail if "@type" not in wsc
        ]


def gen_field_doc(field: Field, doc: str | None) -> str:
    type_repr = f"@type: {inspect.formatannotation(field.type)}"
    return f"{doc}\n\n{type_repr}" if doc else type_repr


def format_exist(
    fields: dict[str, tuple[Field, str | None]],
    container: JObject,
) -> None:
    for k, v in container.items():
        if k in fields:
            field, doc = fields.pop(k)
            k: JString = convert(k)
            container[k] = convert(v)
            remove_generated_comment(k)
            remove_generated_comment(v)
            k.json_before.append(BlockStyleComment(gen_field_doc(field, doc)))


def format_not_exist(
    fields: dict[str, tuple[Field, str | None]],
    container: JObject,
) -> None:
    remove_generated_comment(container)
    for k, (field, doc) in fields.items():
        k = convert(k)
        if field.default_factory is not MISSING:
            default = field.default_factory()
        else:
            default = field.default
        if is_dataclass(default):
            v = asdict(default)
        elif default is MISSING:
            v = None
        else:
            v = default
        container[k] = convert(v)
        k.json_before.append(BlockStyleComment(gen_field_doc(field, doc)))


def format_with_model(container: JObject, model: type[ConfigModel]) -> None:
    if not isinstance(container, JObject):
        raise TypeError(f"{container} is not a json object.")

    fields = {
        k: (f, cast(Union[str, None], f.metadata.get("description")))
        for k, f in model.__dataclass_fields__.items()
    }
    format_exist(fields, container)
    format_not_exist(fields, container)



from dataclasses import dataclass, field

from kayaku.doc_parse import store_field_description


def test_extract_attr_docs():
    @dataclass
    class M:
        a: int = 5  # undocumented
        b: int = field(default=6)
        """b document"""
        c: int = 7
        """long
        long
        c document
        """

    store_field_description(M, M.__dataclass_fields__)
    assert {
        k: f.metadata.get("description") for k, f in M.__dataclass_fields__.items()
    } == {
        "a": None,
        "b": "b document",
        "c": "long\nlong\nc document",
    }



"""Prettify a JSON Container.
"""
from __future__ import annotations

import inspect
from typing import Literal as Constant
from typing import TypeVar

import regex

from .backend.types import (
    WSC,
    Array,
    BlockStyleComment,
    Comment,
    Identifier,
    JObject,
    JString,
    JType,
    Quote,
    WhiteSpace,
    convert,
)

T_Container = TypeVar("T_Container", Array, JObject)

# LINK: https://262.ecma-international.org/5.1/#sec-7.6
IDENTIFIER_PATTERN = regex.compile(
    r"([\p{Lu}\p{Ll}\p{Lt}\p{Lm}\p{Lo}\p{Nl}$_]|\\u[0-9a-fA-F]{4})([\p{Lu}\p{Ll}\p{Lt}\p{Lm}\p{Lo}\p{Nl}$_\p{Mn}\p{Mc}\p{Nd}\p{Pc}\u200C\u200D]|\\u[0-9a-fA-F]{4})*".replace(
        r"\u200C", "\u200C"
    ).replace(
        r"\u200D", "\u200D"
    )
)


class Prettifier:
    """容器格式化工具"""

    def __init__(
        self,
        indent: int = 4,
        trail_comma: bool = False,
        key_quote: Quote | None | Constant[False] = Quote.DOUBLE,
        string_quote: Quote | None = Quote.DOUBLE,
        unfold_single: bool = False,
    ):
        """

        Args:
            indent (int, optional): 缩进数量. Defaults to 4.
            trail_comma (bool, optional): 是否要为容器增加尾随逗号. Defaults to False.
            key_quote (Quote | None | Constant[False], optional): 键使用的引号风格,
                None 为保留, Quote.DOUBLE 为双引号, Quote.SINGLE 为单引号, False 为尽可能去除引号.
                Defaults to Quote.DOUBLE.
            string_quote (Quote | None, optional): 值中字符串使用的引号风格, 解释同上. Defaults to Quote.DOUBLE.
            unfold_single (bool, optional): 单个元素的容器是否展开. Defaults to False.
        """
        self.indent: int = indent
        self.trail_comma: bool = trail_comma
        self.key_quote: Quote | Constant[False] | None = key_quote
        self.string_quote: Quote | None = string_quote
        self.unfold_single: bool = unfold_single
        self.layer: int = 0

    @staticmethod
    def collect_comments(obj: JType) -> list[Comment]:
        return [i for i in obj.json_before + obj.json_after if isinstance(i, Comment)]

    @staticmethod
    def require_newline(wsc_list: list[WSC]) -> bool:
        res: bool = True
        for t in wsc_list:
            if isinstance(t, Comment):
                res |= isinstance(t, BlockStyleComment)
                break
            if "\n" in t:
                return True
            res = False
        return res

    @staticmethod
    def clean_comment(lines: list[str]) -> list[str]:
        res = []
        if lines[-1].strip() == "":
            del lines[-1]
        for i in map(str.lstrip, lines):
            if i.startswith("*"):
                i = i[1:]
            if i.startswith(" "):
                i = i[1:]
            res.append(i)
        return res

    def gen_comment_block(self, comment: str) -> BlockStyleComment:
        lines: list[str] = inspect.cleandoc(comment).splitlines()
        if len(lines) <= 1:
            return BlockStyleComment(comment)
        lines = self.clean_comment(lines)
        indentation: str = " " * self.layer * self.indent
        return BlockStyleComment(
            "".join(
                f"\n{indentation} * {i}" if i else f"\n{indentation} *" for i in lines
            )
            + f"\n{indentation} "
        )

    def format_container(self, new_obj: T_Container, obj: T_Container) -> T_Container:
        comments = [i for i in obj.json_container_tail if isinstance(i, Comment)]
        newline = Prettifier.require_newline(obj.json_container_tail)
        new_obj.json_container_tail = self.format_wsc(comments, newline)
        return new_obj

    def format_wsc(self, comments: list[Comment], require_newline: bool) -> list[WSC]:
        indent: str = " " * self.layer * self.indent
        wsc_list: list[WSC] = []
        for comment in comments:
            wsc_list.extend(
                [
                    WhiteSpace(f"\n{indent}"),
                    self.gen_comment_block(comment)
                    if isinstance(comment, BlockStyleComment)
                    else comment,
                ]
            )
        if wsc_list and not require_newline:
            wsc_list[0] = WhiteSpace(" ")
        wsc_list.append(WhiteSpace(f"\n{indent}"))
        return wsc_list

    def convert_key(self, obj: JString | Identifier | str) -> JString | Identifier:
        if self.key_quote is None:
            return convert(obj)
        if self.key_quote:
            string = JString(obj).__post_init__(quote=self.key_quote)
        else:
            string = (
                Identifier(obj)
                if IDENTIFIER_PATTERN.fullmatch(obj)
                else JString(obj).__post_init__(Quote.DOUBLE)
            )
        obj = convert(obj)
        string.json_before = obj.json_before
        string.json_after = obj.json_after
        return string

    def prettify_object(self, obj: JObject) -> JObject:
        """Prettify a JSON Object."""
        # Object(*(KVPair ,) container_trail)
        # KVPair(key : value)
        if not obj and not obj.json_container_tail:
            return JObject().__post_init__()
        if len(obj) == 1 and not self.unfold_single:
            k, v = next(iter(obj.items()))
            k, v = self.convert_key(k), convert(v)
            sub_comments = self.collect_comments(k) + self.collect_comments(v)
            if not (
                sub_comments or (v and isinstance(v, (list, tuple, dict)))
            ):  # is simple type
                k.__json_clear__()
                v.__json_clear__()
                v.json_before.append(WhiteSpace(" "))
                return JObject({k: v}).__post_init__()
        self.layer += 1
        new_obj = JObject().__post_init__()
        # Preserve container tail
        if obj and not obj.json_container_trailing_comma and (pair := obj.popitem()):
            k, v = self.convert_key(pair[0]), convert(pair[1])
            self.swap_tail(v, obj)
            obj[k] = v

        for k, v in obj.items():
            k, v = self.convert_key(k), convert(v)
            sub_comments = self.collect_comments(k) + self.collect_comments(v)
            newline = self.require_newline(k.json_before)
            k.__json_clear__()
            v.__json_clear__()
            if isinstance(v, (Array, JObject)):
                v = self.prettify(v)
            v.json_before.append(WhiteSpace(" "))
            new_obj[k] = v
            k.json_before = self.format_wsc(sub_comments, newline)
        self.layer -= 1
        return self.format_container(new_obj, obj)

    def prettify_array(self, arr: Array) -> Array:
        """Prettify a JSON Array."""
        # Array(*(value ,) container_trail)
        if not arr and not arr.json_container_tail:
            return Array().__post_init__()
        if len(arr) == 1 and not self.unfold_single:
            v: JType = convert(arr[0])
            sub_comments = self.collect_comments(v)
            if not (
                sub_comments or (v and isinstance(v, (list, tuple, dict)))
            ):  # is simple type
                v.__json_clear__()
                return Array((v,)).__post_init__()
        new_arr = Array().__post_init__()
        self.layer += 1
        # Preserve container tail
        if arr and not arr.json_container_trailing_comma:
            v = convert(arr.pop())
            self.swap_tail(v, arr)
            arr.append(v)
        for v in arr:
            v: JType = convert(v)
            sub_comments = self.collect_comments(v)
            newline = self.require_newline(v.json_before)
            v.__json_clear__()
            if isinstance(v, (Array, JObject)):
                v = self.prettify(v)
            v.json_before = self.format_wsc(sub_comments, newline)
            new_arr.append(v)
        self.layer -= 1
        return self.format_container(new_arr, arr)

    def swap_tail(self, v: JType, obj: Array | JObject) -> None:
        if any(isinstance(wsc, Comment) for wsc in v.json_after):
            obj.json_container_tail = v.json_after
            v.json_after = []

    def prettify(self, container: JObject | Array) -> JObject | Array:
        return (
            self.prettify_object(container)
            if isinstance(container, JObject)
            else self.prettify_array(container)
        )



"""Uniformed JSON backend for JSON5.

Ported from <https://github.com/noirbizarre/json4humans> under the MIT license.
"""
from __future__ import annotations

import io
from pathlib import Path
from typing import Any, TextIO

from lark.lark import Lark

from .encode import Encoder as Encoder
from .env import DEBUG
from .transform import transformer

parser = Lark.open(
    "grammar/json5.lark",
    rel_to=__file__,
    lexer="auto",
    parser="lalr",
    start="value",
    maybe_placeholders=False,
    regex=True,
    transformer=transformer,
)

debug_parser = Lark.open(
    "grammar/json5.lark",
    rel_to=__file__,
    lexer="auto",
    parser="lalr",
    start="value",
    maybe_placeholders=False,
    regex=True,
    transformer=None,
)


def loads(src: str) -> Any:
    """
    Parse JSON from a string
    """
    if not DEBUG.get():
        return parser.parse(src)
    tree = debug_parser.parse(src)
    return transformer.transform(tree)


def load(file: TextIO | Path) -> Any:
    """
    Parse JSON from a file-like object
    """
    data = file.read_text() if isinstance(file, Path) else file.read()
    return loads(data)


def dumps(obj: Any, encoder_cls: type[Encoder] = Encoder, endline: bool = False) -> str:
    """
    Serialize JSON to a string
    """
    fp = io.StringIO()
    dump(obj, fp, encoder_cls, endline)
    return fp.getvalue()


def dump(
    obj: Any,
    out: TextIO | Path,
    encoder_cls: type[Encoder] = Encoder,
    endline: bool = False,
) -> None:
    """
    Serialize JSON to a file-like object
    """
    out = out.open("w") if isinstance(out, Path) else out
    encoder_cls(out).encode(obj)
    if endline:
        out.write("\n")



from __future__ import annotations

import enum
import re
from dataclasses import fields, field
from datetime import date, datetime, time
import sys
import types
from typing import Any, Mapping, Sequence, TypeVar, Union, TYPE_CHECKING

import typing_extensions
from dacite.config import Config
from dacite.core import from_dict as _from_dict
from typing_extensions import TypeAlias

from .backend.types import Array, JContainer, JObject, JType, JWrapper, convert
from .schema_gen import ConfigModel

DomainType: TypeAlias = "tuple[str, ...]"

T = TypeVar("T", bound=ConfigModel)

Unions = {Union,} if sys.version_info < (3, 10) else {Union, types.UnionType}


def copy_meta(src: Any, dst: JType):
    if isinstance(dst, JContainer):
        dst.json_container_tail = getattr(
            src, "json_container_tail", dst.json_container_tail
        )
    dst.json_before = getattr(src, "json_before", dst.json_before)
    dst.json_after = getattr(src, "json_after", dst.json_after)


def _update_array(container: Array, data: list):
    for i in range(len(container)):
        val = container[i]
        if isinstance(val, (ConfigModel, dict)):
            new_container = JObject()
            update(new_container, val, delete=True)
            val = new_container
        elif isinstance(val, (re.Pattern, date, datetime, time)):
            val = convert(
                val.pattern if isinstance(val, re.Pattern) else val.isoformat()
            )
        elif isinstance(val, enum.Enum):
            val = convert(val.value)
        elif isinstance(val, Sequence) and not isinstance(val, str):
            val: JType = convert(val if isinstance(val, list) else list(val))
            _update_array(val, [])
        else:
            val = convert(val)

        if i < len(data) and isinstance(data[i], JType):
            copy_meta(data[i], val)
        container[i] = val
    if len(container) < len(data):
        container.json_container_trailing_comma = True


def update(container: JObject, data: ConfigModel | dict, delete: bool = False):
    if isinstance(data, ConfigModel):
        k_v_pairs = {f.name: getattr(data, f.name) for f in fields(data)}
    else:
        k_v_pairs = data
    to_be_popped: set[str] = set(container.keys() if delete else ())
    for k, v in k_v_pairs.items():
        k = convert(k)
        to_be_popped.discard(k)
        origin_v = container.get(k, None)
        if isinstance(v, (ConfigModel, dict)):
            new_v = container.setdefault(k, JObject())
            update(new_v, v, delete=True)
            v = new_v
        elif isinstance(v, (re.Pattern, date, datetime, time)):
            v = convert(v.pattern if isinstance(v, re.Pattern) else v.isoformat())
        elif isinstance(v, enum.Enum):
            v = convert(v.value)
        elif isinstance(v, Sequence) and not isinstance(v, str):
            v: JType = convert(v if isinstance(v, list) else list(v))
            _update_array(v, origin_v or [])
        else:
            v: JType = convert(v)
        if origin_v is not None:
            copy_meta(origin_v, v)
        container[k] = v
    for k in to_be_popped:
        container.pop(k, None)


class _KayakuDaciteTypeHook(dict):
    def __init__(self):
        super().__init__(
            {
                datetime: self.hook_datetime,
                time: self.hook_time,
                date: self.hook_date,
                re.Pattern: self.hook_re_pattern,
                bool: self.hook_bool,
                type(None): self.hook_none,
            }
        )

    def hook_datetime(self, v: str | datetime) -> datetime:
        return datetime.fromisoformat(v) if isinstance(v, str) else v

    def hook_time(self, v: str | time) -> time:
        return time.fromisoformat(v) if isinstance(v, str) else v

    def hook_date(self, v: str | date) -> date:
        return date.fromisoformat(v) if isinstance(v, str) else v

    def hook_re_pattern(self, v: str | re.Pattern):
        return re.compile(v) if isinstance(v, str) else v

    def hook_bool(self, v: JWrapper[bool] | bool) -> bool:
        return v.value if isinstance(v, JWrapper) else v

    def hook_none(self, v: JWrapper[None] | None) -> None:
        return v.value if isinstance(v, JWrapper) else v

    def __contains__(self, o: object) -> bool:
        if typing_extensions.get_origin(o) == Union:
            return any(
                dict.__contains__(self, arg) for arg in typing_extensions.get_args(o)
            )
        return super().__contains__(o)

    def __getitem__(self, k: Any) -> Any:
        if typing_extensions.get_origin(k) in Unions:

            def applier(arg):
                for func in [
                    self[arg]
                    for arg in typing_extensions.get_args(k)
                    if dict.__contains__(self, arg)
                ]:
                    arg = func(arg)
                return arg

            return applier
        return super().__getitem__(k)


_TYPE_HOOK = _KayakuDaciteTypeHook()


def from_dict(model: type[T], data: dict[str, Any]) -> T:
    return _from_dict(
        model,
        data,
        Config(
            type_hooks=_TYPE_HOOK,
            cast=[enum.Enum],
        ),
    )


if TYPE_CHECKING:

    def copying_field_stub(
        default: T,
        *,
        init: bool = True,
        repr: bool = True,
        hash: bool | None = None,
        compare: bool = True,
        metadata: Mapping[Any, Any] | None = None,
        kw_only: bool = False,
    ) -> T:
        """自动 deepcopy 默认值的 `dataclass.field`.

        Args:
            default (T): The default value for this field, which will be deepcopied on each call.
            init (bool, optional): If true (the default), this field is included in `__init__()`.
            repr (bool, optional): If true (the default), this field is included in `repr()`.
            hash (bool | None, optional): If true, this field is included in the generated `__hash__()` method. If None (the default), use the value of compare.
            compare (bool, optional): If true (the default), this field is included in comparison methods.
            metadata (Mapping[Any, Any] | None, optional): A mapping which is stored but not otherwise examined by dataclass.
            kw_only (bool, optional): If true, this field will be marked as keyword-only. (3.10+)

        Returns:
            T: The field object, faked to be the same type of `default` in typ checkers' eyes.
        """
        ...

else:

    def copying_field_impl(default, **kwargs):
        from copy import deepcopy

        default_value = deepcopy(default)
        return field(default_factory=lambda: deepcopy(default_value), **kwargs)


copying_field = copying_field_stub if TYPE_CHECKING else copying_field_impl



from __future__ import annotations

from contextvars import ContextVar

from .spec import FormattedPath, PathSpec, SourceSpec


class _SuffixNode:
    bound: tuple[SourceSpec, PathSpec] | None

    def __init__(self) -> None:
        self.bound = None
        self.nxt: dict[str, _SuffixNode] = {}

    def insert(self, frags: list[str], index: int = 0) -> _SuffixNode:
        if index == len(frags):
            return self
        nxt = self.nxt.setdefault(frags[index], _SuffixNode())
        return nxt.insert(frags, index + 1)

    def lookup(
        self, frags: list[str], index: int = 0
    ) -> tuple[int, tuple[SourceSpec, PathSpec] | None]:
        res = index, self.bound
        if index < len(frags):
            if nxt_nd := self.nxt.get(frags[index], None):
                if (lookup_res := nxt_nd.lookup(frags, index + 1)) and lookup_res[1]:
                    return lookup_res
        return res


class _PrefixNode:
    suffix: _SuffixNode | None

    def __init__(self) -> None:
        self.suffix = None
        self.nxt: dict[str, _PrefixNode] = {}

    def insert(self, frags: list[str], index: int = 0) -> _SuffixNode:
        if index == len(frags):
            if not self.suffix:
                self.suffix = _SuffixNode()
            return self.suffix
        nxt = self.nxt.setdefault(frags[index], _PrefixNode())
        return nxt.insert(frags, index + 1)

    def lookup(
        self, frags: list[str], index: int = 0
    ) -> tuple[tuple[SourceSpec, PathSpec], FormattedPath] | None:
        if index < len(frags) and (nxt_nd := self.nxt.get(frags[index], None)):
            if lookup_res := nxt_nd.lookup(frags, index + 1):
                return lookup_res
        if self.suffix:
            suffix_ind, spec = self.suffix.lookup(list(reversed(frags[index:])))
            if spec:
                src_spec, path_spec = spec
                parts = (
                    src_spec.section.prefix
                    + frags[index : -suffix_ind or None]
                    + src_spec.section.suffix
                )
                if formatted := path_spec.format(parts):
                    return spec, formatted


def insert(src: SourceSpec, path: PathSpec) -> None:
    prefix, suffix = src.prefix, src.suffix
    target_nd = _root.get().insert(prefix).insert(list(reversed(suffix)))
    if target_nd.bound:
        raise ValueError(
            f"{'.'.join(prefix + ['*'] + suffix)} is already bound to {target_nd.bound}"
        )
    target_nd.bound = (src, path)


def lookup(domains: list[str]) -> FormattedPath:
    if res := _root.get().lookup(domains):
        return res[1]
    raise ValueError(f"Unable to lookup {'.'.join(domains)}")


_root: ContextVar[_PrefixNode] = ContextVar("kayaku.__root__", default=_PrefixNode())



import io

import pytest

from kayaku import backend
from kayaku.backend.types import Comment, JWrapper, WhiteSpace, convert
from kayaku.backend.wsc import encode_wsc

test_input = """\
{
    "base":{
        /*
        * More annotations
        *
        * @type: int
        */
        "account": 0,
        "multi": "abc\\
def\\
lll\\
",
        'pragma': '中文。\\
asd'
    },
    "next":{
        /*@type: List[dict]*/
        "dct":[
            {"3": 5}
        ],
        /*
        * Test annotating
        * What about double line
        *
        * @type: Optional[P]
        */
        "p": null,
        "q": true, // attempt
        "v": false, /* wow */
        "s": "trail",
        "mult": "abc\\
def",
    },
    identifier: 'SINGLE',
    numbers: [+NaN, -Infinity, 0x7b, +0.5e10, .777, 777.],
    "$schema": "file:///F:/PythonProjects/Graiax/kayaku/config.schema.json"
}
"""


def test_backend_round_trip():
    import kayaku.backend.env
    from kayaku import backend

    in_io = io.StringIO(test_input)
    out_io = io.StringIO()

    assert backend.dumps(backend.loads(test_input)) == test_input
    obj = backend.load(in_io)
    assert obj["base"]["multi"] == "abcdeflll"
    assert obj["base"]["pragma"] == "中文。asd"
    backend.dump(obj, out_io)
    assert out_io.getvalue() == test_input

    backend.env.DEBUG.set(True)

    in_io = io.StringIO(test_input)
    out_io = io.StringIO()

    assert backend.dumps(backend.loads(test_input)) == test_input
    backend.dump(backend.load(in_io), out_io)
    assert out_io.getvalue() == test_input


def test_round_trip_raw():
    commented = "/*abc*/ 5"
    assert backend.dumps(backend.loads(commented)) == commented


def test_load_repr():
    assert str(backend.loads("123")) == "123"
    assert str(backend.loads("0x123")) == "0x123"
    assert backend.loads("null") == None
    assert backend.loads("true") == True
    assert {None: 5}[backend.loads("null")]


def test_dump_val():
    assert backend.dumps(123) == "123"
    assert backend.dumps(True) == "true"
    assert backend.dumps(False) == "false"
    assert backend.dumps("plural") == '"plural"'
    float_val = backend.loads("0.5e10")
    float_val.origin = "123"
    assert backend.loads(backend.dumps(float_val)) == float_val
    assert backend.loads(backend.dumps(0.5)) == 0.5
    hex_val = backend.loads("0xccc")
    hex_val.origin = "45"
    assert backend.loads(backend.dumps(hex_val)) == hex_val
    with pytest.raises(NotImplementedError):
        backend.dumps(object())
    with pytest.raises(NotImplementedError):
        backend.dumps(JWrapper(NotImplemented))


def test_convert():
    for obj in [1, 1.0, True, False, None, {}, []]:
        assert convert(obj) == obj
    with pytest.raises(TypeError):
        convert(object())


def test_wsc():
    v = backend.loads("/*123*/[]")
    v.__json_clear__()
    assert backend.dumps(v) == "[]"
    assert encode_wsc(WhiteSpace("     \n")) == "     \n"
    with pytest.raises(NotImplementedError):
        encode_wsc(Comment("abstract"))


def test_horrible_quote():
    target_single = r"""{"a": '"\'"\'"\'"\''}"""
    target_double = r"""{"a": "\"'\"'\"'\"'"}"""
    assert backend.loads(target_double) == {"a": """"'"'"'"'"""}
    assert backend.loads(target_single) == {"a": """"'"'"'"'"""}
    assert (
        backend.dumps({"a": """"'"'"'"'"""}) == '{"a":"\\"\'\\"\'\\"\'\\"\'"}'
    )  # default: condense
    assert backend.dumps(backend.loads(target_single)) == target_single
    assert backend.dumps(backend.loads(target_double)) == target_double



"""dataclass schema generator tests.

Modified from https://github.com/Peter554/dc_schema.


MIT License

Copyright (c) 2022 Peter Byfield

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import dataclasses
import datetime
import enum
import re
import typing as t

import pytest
import typing_extensions as t_e
from jsonschema.validators import Draft202012Validator

from kayaku.schema_gen import (
    ConfigModel,
    ContainerSchema,
    NumberSchema,
    Schema,
    SchemaGenerator,
    StringSchema,
)


def get_schema(obj: type[ConfigModel]):
    class NameOnlyGen(SchemaGenerator):
        def retrieve_name(self, typ: t.Type) -> str:
            return typ.__name__

    return NameOnlyGen.from_dc(obj)


@dataclasses.dataclass
class DcPrimitives:
    b: bool
    i: int
    f: float
    s: str


def test_get_schema_primitives():
    schema = get_schema(DcPrimitives)
    print(schema)
    Draft202012Validator.check_schema(schema)
    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcPrimitives",
        "properties": {
            "b": {"type": "boolean"},
            "i": {"type": "integer"},
            "f": {"type": "number"},
            "s": {"type": "string"},
        },
        "required": ["b", "i", "f", "s"],
    }


@dataclasses.dataclass
class DcOptional:
    a: int = 42
    b: int = dataclasses.field(default=42)
    c: int = dataclasses.field(default_factory=lambda: 42)
    d: str = "foo"
    e: bool = False
    f: None = None
    g: float = 1.1
    h: t.Tuple[int, float] = (1, 1.1)


def test_get_schema_optional_fields():
    """optional field === field with a default (!== t.Optional)"""
    schema = get_schema(DcOptional)
    print(schema)
    Draft202012Validator.check_schema(schema)
    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcOptional",
        "properties": {
            "a": {"type": "integer", "default": 42},
            "b": {"type": "integer", "default": 42},
            "c": {"type": "integer"},
            "d": {"type": "string", "default": "foo"},
            "e": {"type": "boolean", "default": False},
            "f": {"type": "null", "default": None},
            "g": {"type": "number", "default": 1.1},
            "h": {
                "type": "array",
                "prefixItems": [{"type": "integer"}, {"type": "number"}],
                "minItems": 2,
                "maxItems": 2,
                "default": [1, 1.1],
            },
        },
    }


@dataclasses.dataclass
class DcUnion:
    a: t.Union[int, str]


def test_get_schema_union():
    schema = get_schema(DcUnion)
    print(schema)
    Draft202012Validator.check_schema(schema)
    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcUnion",
        "properties": {"a": {"anyOf": [{"type": "integer"}, {"type": "string"}]}},
        "required": ["a"],
    }


@dataclasses.dataclass
class DcNone:
    a: None
    b: t.Optional[int]
    c: t.Union[None, int]


def test_get_schema_nullable():
    schema = get_schema(DcNone)
    print(schema)
    Draft202012Validator.check_schema(schema)
    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcNone",
        "properties": {
            "a": {"type": "null"},
            "b": {"anyOf": [{"type": "integer"}, {"type": "null"}]},
            "c": {"anyOf": [{"type": "null"}, {"type": "integer"}]},
        },
        "required": ["a", "b", "c"],
    }


@dataclasses.dataclass
class DcDict:
    a: dict
    b: t.Dict[str, int]


def test_get_schema_dict():
    schema = get_schema(DcDict)
    print(schema)
    Draft202012Validator.check_schema(schema)
    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcDict",
        "properties": {
            "a": {"type": "object"},
            "b": {"type": "object", "additionalProperties": {"type": "integer"}},
        },
        "required": ["a", "b"],
    }


@dataclasses.dataclass
class DcList:
    a: list
    b: t.List[bool]


def test_get_schema_list():
    schema = get_schema(DcList)
    print(schema)
    Draft202012Validator.check_schema(schema)
    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcList",
        "properties": {
            "a": {"type": "array"},
            "b": {"type": "array", "items": {"type": "boolean"}},
        },
        "required": ["a", "b"],
    }


@dataclasses.dataclass
class DcTuple:
    a: tuple
    b: t.Tuple[int, ...]
    c: t.Tuple[int, bool, str]


def test_get_schema_tuple():
    schema = get_schema(DcTuple)
    print(schema)
    Draft202012Validator.check_schema(schema)
    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcTuple",
        "properties": {
            "a": {"type": "array"},
            "b": {"type": "array", "items": {"type": "integer"}},
            "c": {
                "type": "array",
                "prefixItems": [
                    {"type": "integer"},
                    {"type": "boolean"},
                    {"type": "string"},
                ],
                "minItems": 3,
                "maxItems": 3,
            },
        },
        "required": ["a", "b", "c"],
    }


@dataclasses.dataclass
class DcRefsChild:
    c: str


@dataclasses.dataclass
class DcRefs:
    a: DcRefsChild
    b: t.List[DcRefsChild]


def test_get_schema_refs():
    schema = get_schema(DcRefs)
    print(schema)
    Draft202012Validator.check_schema(schema)
    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcRefs",
        "properties": {
            "a": {"$ref": "#/$defs/DcRefsChild"},
            "b": {
                "type": "array",
                "items": {"$ref": "#/$defs/DcRefsChild"},
            },
        },
        "required": ["a", "b"],
        "$defs": {
            "DcRefsChild": {
                "type": "object",
                "title": "DcRefsChild",
                "properties": {"c": {"type": "string"}},
                "required": ["c"],
            }
        },
    }


@dataclasses.dataclass
class DcRefsSelf:
    a: str
    b: t.Optional[DcRefsSelf]
    c: t.List[DcRefsSelf]


def test_get_schema_self_refs():
    schema = get_schema(DcRefsSelf)
    print(schema)
    Draft202012Validator.check_schema(schema)
    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcRefsSelf",
        "properties": {
            "a": {"type": "string"},
            "b": {"anyOf": [{"$ref": "#"}, {"type": "null"}]},
            "c": {"type": "array", "items": {"$ref": "#"}},
        },
        "required": ["a", "b", "c"],
    }


@dataclasses.dataclass
class DcLiteral:
    a: t.Literal[1, "two", 3, None]
    b: t.Literal[42, 43] = 42


def test_get_schema_literal():
    schema = get_schema(DcLiteral)
    Draft202012Validator.check_schema(schema)
    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcLiteral",
        "properties": {
            "a": {"enum": [1, "two", 3, None]},
            "b": {"enum": [42, 43], "default": 42},
        },
        "required": ["a"],
    }


@dataclasses.dataclass
class DcAny:
    a: t.Any
    x: t.Union[int, t.Any] = 4
    b: t_e.Annotated[t.Any, Schema("B!!!")] = 5


def test_get_schema_any():
    schema = get_schema(DcAny)
    Draft202012Validator.check_schema(schema)
    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcAny",
        "properties": {
            "a": {},
            "b": {"default": 5, "title": "B!!!"},
            "x": {"anyOf": [{"type": "integer"}, {}], "default": 4},
        },
        "required": ["a"],
    }


class MyEnum(enum.Enum):
    a = enum.auto()
    b = enum.auto()


@dataclasses.dataclass
class DcEnum:
    a: MyEnum
    b: MyEnum = MyEnum.a


def test_get_schema_enum():
    schema = get_schema(DcEnum)
    print(schema)
    Draft202012Validator.check_schema(schema)
    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcEnum",
        "properties": {
            "a": {"$ref": "#/$defs/MyEnum"},
            "b": {"$ref": "#/$defs/MyEnum", "default": 1},
        },
        "required": ["a"],
        "$defs": {"MyEnum": {"title": "MyEnum", "enum": [1, 2]}},
    }


@dataclasses.dataclass
class DcSet:
    a: set
    b: t.Set[int]


def test_get_schema_set():
    schema = get_schema(DcSet)
    print(schema)
    Draft202012Validator.check_schema(schema)
    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcSet",
        "properties": {
            "a": {"type": "array", "uniqueItems": True},
            "b": {"type": "array", "items": {"type": "integer"}, "uniqueItems": True},
        },
        "required": ["a", "b"],
    }


@dataclasses.dataclass
class DcStrAnnotated:
    a: t_e.Annotated[str, StringSchema(min_length=3, max_length=5)]
    b: t_e.Annotated[str, StringSchema(format="date", pattern=r"^\d.*")] = "2000-01-01"


def test_get_schema_str_annotation():
    schema = get_schema(DcStrAnnotated)
    print(schema)
    Draft202012Validator.check_schema(schema)
    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcStrAnnotated",
        "properties": {
            "a": {"type": "string", "minLength": 3, "maxLength": 5},
            "b": {
                "type": "string",
                "default": "2000-01-01",
                "pattern": "^\\d.*",
                "format": "date",
            },
        },
        "required": ["a"],
    }


@dataclasses.dataclass
class DcNumberAnnotated:
    a: t_e.Annotated[int, NumberSchema(minimum=1, exclusive_maximum=11)]
    b: t.List[t_e.Annotated[int, NumberSchema(minimum=0)]]
    c: t.Optional[t_e.Annotated[int, NumberSchema(minimum=0)]]
    d: t_e.Annotated[
        float, NumberSchema(maximum=12, exclusive_minimum=17, multiple_of=5)
    ] = 33.1


def test_get_schema_number_annotation():
    schema = get_schema(DcNumberAnnotated)
    print(schema)
    Draft202012Validator.check_schema(schema)
    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcNumberAnnotated",
        "properties": {
            "a": {"type": "integer", "minimum": 1, "exclusiveMaximum": 11},
            "b": {"type": "array", "items": {"type": "integer", "minimum": 0}},
            "c": {"anyOf": [{"type": "integer", "minimum": 0}, {"type": "null"}]},
            "d": {
                "type": "number",
                "default": 33.1,
                "maximum": 12,
                "exclusiveMinimum": 17,
                "multipleOf": 5,
            },
        },
        "required": ["a", "b", "c"],
    }


@dataclasses.dataclass
class DcDateTime:
    a: datetime.datetime
    b: datetime.date


def test_get_schema_date_time():
    schema = get_schema(DcDateTime)
    print(schema)
    Draft202012Validator.check_schema(schema)
    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcDateTime",
        "properties": {
            "a": {"type": "string", "format": "date-time"},
            "b": {"type": "string", "format": "date"},
        },
        "required": ["a", "b"],
    }


@dataclasses.dataclass
class DcRegex:
    a: re.Pattern


def test_get_schema_regex():
    schema = get_schema(DcRegex)
    print(schema)
    Draft202012Validator.check_schema(schema)
    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcRegex",
        "properties": {
            "a": {"type": "string", "format": "regex"},
        },
        "required": ["a"],
    }


@dataclasses.dataclass
class DcNotImplemented:
    a: type(NotImplemented)


def test_not_implemented():
    with pytest.raises(NotImplementedError):
        get_schema(DcNotImplemented)


def test_get_schema_regex():
    schema = get_schema(DcRegex)
    print(schema)
    Draft202012Validator.check_schema(schema)
    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcRegex",
        "properties": {
            "a": {"type": "string", "format": "regex"},
        },
        "required": ["a"],
    }


@dataclasses.dataclass
class DcAnnotatedBook:
    title: t_e.Annotated[str, Schema(title="Title")]


class DcAnnotatedAuthorHobby(enum.Enum):
    CHESS = "chess"
    SOCCER = "soccer"


@dataclasses.dataclass
class DcAnnotatedAuthor:
    name: t_e.Annotated[
        str,
        Schema(description="the name of the author", examples=["paul", "alice"]),
    ]
    books: t_e.Annotated[
        t.List[DcAnnotatedBook],
        Schema(description="all the books the author has written"),
    ]
    hobby: t_e.Annotated[DcAnnotatedAuthorHobby, Schema(deprecated=True)]
    age: t_e.Annotated[t.Union[int, float], Schema(description="age in years")] = 42


def test_config_model_abc():
    @dataclasses.dataclass
    class C:
        a: int

    assert isinstance(C(5), ConfigModel)
    assert issubclass(C, ConfigModel)


def test_get_schema_annotation():
    schema = get_schema(DcAnnotatedAuthor)
    print(schema)
    Draft202012Validator.check_schema(schema)
    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcAnnotatedAuthor",
        "properties": {
            "name": {
                "type": "string",
                "description": "the name of the author",
                "examples": ["paul", "alice"],
            },
            "books": {
                "type": "array",
                "items": {"$ref": "#/$defs/DcAnnotatedBook"},
                "description": "all the books the author has written",
            },
            "hobby": {
                "$ref": "#/$defs/DcAnnotatedAuthorHobby",
                "deprecated": True,
            },
            "age": {
                "anyOf": [{"type": "integer"}, {"type": "number"}],
                "default": 42,
                "description": "age in years",
            },
        },
        "required": ["name", "books", "hobby"],
        "$defs": {
            "DcAnnotatedBook": {
                "type": "object",
                "title": "DcAnnotatedBook",
                "properties": {"title": {"type": "string", "title": "Title"}},
                "required": ["title"],
            },
            "DcAnnotatedAuthorHobby": {
                "title": "DcAnnotatedAuthorHobby",
                "enum": ["chess", "soccer"],
            },
        },
    }


@dataclasses.dataclass
class DcSchemaConfigChild:
    a: int


@dataclasses.dataclass
class DcSchemaConfig:
    a: str
    child_1: DcSchemaConfigChild
    child_2: t_e.Annotated[DcSchemaConfigChild, Schema(title="2nd child")]
    friend: t_e.Annotated[DcSchemaConfig, Schema(title="a friend")]


def test_get_schema_config():
    schema = get_schema(DcSchemaConfig)
    print(schema)
    Draft202012Validator.check_schema(schema)
    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcSchemaConfig",
        "properties": {
            "a": {"type": "string"},
            "child_1": {"$ref": "#/$defs/DcSchemaConfigChild"},
            "child_2": {
                "$ref": "#/$defs/DcSchemaConfigChild",
                "title": "2nd child",
            },
            "friend": {"$ref": "#", "title": "a friend"},
        },
        "required": ["a", "child_1", "child_2", "friend"],
        "$defs": {
            "DcSchemaConfigChild": {
                "type": "object",
                "title": "DcSchemaConfigChild",
                "properties": {"a": {"type": "integer"}},
                "required": ["a"],
            }
        },
    }


@dataclasses.dataclass
class DcListAnnotation:
    a: t_e.Annotated[
        t.List[int],
        ContainerSchema(min_items=3, max_items=5, unique_items=True),
    ]
    b: t_e.Annotated[
        t.Tuple[float, ...], ContainerSchema(min_items=3, max_items=10)
    ] = ()


def test_get_schema_list_annotation():
    schema = get_schema(DcListAnnotation)
    print(schema)
    Draft202012Validator.check_schema(schema)
    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcListAnnotation",
        "properties": {
            "a": {
                "type": "array",
                "items": {"type": "integer"},
                "minItems": 3,
                "maxItems": 5,
                "uniqueItems": True,
            },
            "b": {
                "type": "array",
                "items": {"type": "number"},
                "default": [],
                "minItems": 3,
                "maxItems": 10,
            },
        },
        "required": ["a"],
    }


@dataclasses.dataclass
class DcIncorrectStringAnnotation:
    price: t_e.Annotated[int, StringSchema(pattern=r"\d+")]


@dataclasses.dataclass
class DcIncorrectNumberAnnotation:
    price: t_e.Annotated[str, NumberSchema(maximum=10)]


@dataclasses.dataclass
class DcIncorrectContainerAnnotation:
    price: t_e.Annotated[int, ContainerSchema(max_items=5)]


def test_get_incorrect_annotation():
    for typ in (
        DcIncorrectContainerAnnotation,
        DcIncorrectNumberAnnotation,
        DcIncorrectStringAnnotation,
    ):
        with pytest.raises(TypeError):
            get_schema(typ)


@dataclasses.dataclass
class DcProduct:
    price: t_e.Annotated[float, NumberSchema(minimum=0, maximum=5000)]
    """Price of the product."""

    name: t_e.Annotated[str, Schema(description="Name of the product.")]
    """_name of the product_"""

    category: str
    """Category of the product."""


def test_dc_docstring_merge():
    assert get_schema(DcProduct) == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "object",
        "title": "DcProduct",
        "properties": {
            "price": {
                "type": "number",
                "minimum": 0,
                "maximum": 5000,
                "description": "Price of the product.",
            },
            "name": {
                "type": "string",
                "description": "Name of the product.",
            },
            "category": {
                "type": "string",
                "description": "Category of the product.",
            },
        },
        "required": ["price", "name", "category"],
    }


class Book(t_e.TypedDict):
    name: t_e.Required[str]
    """Book Name"""
    isbn: str
    """ISBN Serial"""
    author: t_e.Annotated[t_e.Required[str], Schema(title="Book Author")]
    """Book Author"""
    year: t_e.Annotated[t_e.NotRequired[int], NumberSchema(maximum=9999)]
    """Publish Year"""


class Disc(t_e.TypedDict, total=False):
    name: t_e.Required[str]
    artists: t.List[str]
    songs: t.List[str]


@dataclasses.dataclass
class DcShelf:
    items: t.List[t.Union[Book, Disc]]


def test_dc_gen_typed_dict():
    assert get_schema(DcShelf) == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$defs": {
            "Book": {
                "properties": {
                    "author": {
                        "description": "Book Author",
                        "title": "Book Author",
                        "type": "string",
                    },
                    "isbn": {"description": "ISBN Serial", "type": "string"},
                    "name": {"description": "Book Name", "type": "string"},
                    "year": {
                        "description": "Publish Year",
                        "maximum": 9999,
                        "type": "integer",
                    },
                },
                "required": ["name", "isbn", "author"],
                "title": "Book",
                "type": "object",
            },
            "Disc": {
                "properties": {
                    "artists": {"items": {"type": "string"}, "type": "array"},
                    "name": {"type": "string"},
                    "songs": {"items": {"type": "string"}, "type": "array"},
                },
                "required": ["name"],
                "title": "Disc",
                "type": "object",
            },
        },
        "properties": {
            "items": {
                "items": {
                    "anyOf": [
                        {"$ref": "#/$defs/Book"},
                        {"$ref": "#/$defs/Disc"},
                    ]
                },
                "type": "array",
            }
        },
        "required": ["items"],
        "title": "DcShelf",
        "type": "object",
    }


class StoreItem(t_e.TypedDict):
    """Represents an item in store."""

    name: str
    """Item Name"""
    price: float
    """Item Price"""


@dataclasses.dataclass
class DcStore:
    """Represents a store."""

    items: t.List[StoreItem]


def test_dc_gen_with_cls_doc():
    assert get_schema(DcStore) == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$defs": {
            "StoreItem": {
                "properties": {
                    "name": {"description": "Item Name", "type": "string"},
                    "price": {"description": "Item Price", "type": "number"},
                },
                "required": ["name", "price"],
                "title": "StoreItem",
                "type": "object",
            },
        },
        "properties": {
            "items": {
                "items": {
                    "$ref": "#/$defs/StoreItem",
                },
                "type": "array",
            }
        },
        "required": ["items"],
        "title": "DcStore",
        "description": "Represents a store.",
        "type": "object",
    }


def test_dc_union_type():
    import sys

    if sys.version_info <= (3, 10):
        pytest.skip("types.UnionType is not available under Python 3.10.")

    @dataclasses.dataclass
    class DcOptional:
        """Represents an optional item."""

        item: int | None = None

    assert get_schema(DcOptional) == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "description": "Represents an optional item.",
        "properties": {
            "item": {"anyOf": [{"type": "integer"}, {"type": "null"}], "default": None}
        },
        "title": "DcOptional",
        "type": "object",
    }



import shutil
import sys
from contextlib import contextmanager
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import List, Union

import pytest

import kayaku
import kayaku.domain
import kayaku.storage
from kayaku import bootstrap, config, create, initialize, save, save_all
from kayaku.backend import loads
from kayaku.pretty import Prettifier
from kayaku.schema_gen import SchemaGenerator


@contextmanager
def start_over():
    base_pth = Path("./temp/full_test").resolve()
    base_pth.mkdir(parents=True, exist_ok=True)
    kayaku.domain._store = kayaku.domain._GlobalStore(Prettifier, SchemaGenerator)
    kayaku.storage._root.set(kayaku.storage._PrefixNode())
    kayaku.initialize({"{**}": base_pth.as_posix() + "/{}::{**}"})
    yield
    shutil.rmtree(base_pth.as_posix())


def test_main():

    start_before_init()
    empty_location_report()

    test_cases = [
        init_fail,
        field_name_collision,
        bootstrap_fail,
        nested_model,
        repl_test,
        basic_test,
        occupy_test,
        reload_test,
        invalid_conf_cls,
    ]
    for case in test_cases:
        with start_over():
            case()


def start_before_init():
    with pytest.raises(RuntimeError):

        @config("config")
        class Conf:
            a: int


def invalid_conf_cls():
    with pytest.raises(ValueError):

        @config("")
        class Conf:
            a: int

    @dataclass
    class InvalidConfigModel:
        a: int

    with pytest.raises(TypeError):
        create(InvalidConfigModel)


def empty_location_report():
    kayaku.domain._store = kayaku.domain._GlobalStore(Prettifier, SchemaGenerator)
    with pytest.raises(ValueError):

        @config("domain.conf.1")
        class Conf:
            a: int


def init_fail():
    with pytest.raises(ValueError):
        kayaku.initialize({"": ""})


def field_name_collision():
    kayaku.initialize(
        {
            "sub.a.{**}": "./temp/full_test/collision",
            "sub.b.{**}": "./temp/full_test/collision",
        }
    )

    @config("sub.a")
    class A:
        a: int

    with pytest.raises(NameError):

        @config("sub.b")
        class B:
            a: int


def bootstrap_fail():
    @config("account")
    class AccountConf:
        account: Union[int, str]
        """Account, int for UserID, str for username."""
        password: str

    with pytest.raises(ValueError):
        bootstrap()


def nested_model():
    @dataclass
    class Sub1:
        a: int = 5
        b: int = 6

    @config("conf.nested.bootstrap")
    class Parent:
        sub1: Sub1 = field(default_factory=Sub1)
        f: int = 5

    @config("conf.nested.other")
    class Other:
        important: bool = True

    save(Other)
    save(create(Other))

    bootstrap()
    save_all()
    assert loads(Path("./temp/full_test/conf.jsonc").read_text()) == {
        "nested": {
            "bootstrap": {
                "f": 5,
                "sub1": {
                    "a": 5,
                    "b": 6,
                },
            },
            "other": {
                "important": True,
            },
        },
        "$schema": Path("./temp/full_test/conf.schema.json").resolve().as_uri(),
    }


def basic_test():
    account_pth = Path("./temp/full_test/account.jsonc").resolve()

    account_pth.touch(exist_ok=True)

    account_pth.write_text("""{"account": "admin@graia", "password": "login_weak"}""")

    @config("account")
    class AccountConf:
        account: Union[int, str]
        """Account, int for UserID, str for username."""
        password: str = ""

    create(AccountConf)

    sub_pth = Path("./temp/full_test/subscription.jsonc")

    sub_pth.write_text(
        """
{
    "organizations": ["GraiaProject", "GraiaCommunity"],
}
"""
    )

    @config("subscription")
    class SubConf:
        organizations: List[str] = field(default_factory=list)
        """Organizations that you want to subscribe (tracks *all* repo events)"""

        repos: List[str] = field(default_factory=list)
        """Repositories that you want to subscribe
        Please avoid duplicating the organizations that you've subscribed.
        """

    save_all()

    bootstrap()
    account = create(AccountConf)
    account_copy = create(AccountConf)
    assert account is account_copy
    assert account.account == "admin@graia"
    assert account.password == "login_weak"

    sub_conf = create(SubConf)
    save_all()

    sub_conf.repos.append("GreyElaina/richuru")
    save(SubConf)
    sub_conf = create(SubConf, flush=True)
    assert sub_conf.repos == ["GreyElaina/richuru"]

    assert (
        sub_pth.read_text()
        == f"""\
{{
    /*
     * Organizations that you want to subscribe (tracks *all* repo events)
     *
     * @type: List[str]
     */
    "organizations": [
        "GraiaProject",
        "GraiaCommunity"
    ],
    /*
     * Repositories that you want to subscribe
     * Please avoid duplicating the organizations that you've subscribed.
     *
     * @type: List[str]
     */
    "repos": ["GreyElaina/richuru"],
    "$schema": "{sub_pth.with_name("subscription.schema.json").resolve().as_uri()}"
}}
"""
    )

    assert (
        account_pth.read_text()
        == f"""\
{{
    /*
     * Account, int for UserID, str for username.
     *
     * @type: Union[int, str]
     */
    "account": "admin@graia",
    /*@type: str*/
    "password": "login_weak",
    "$schema": "{account_pth.with_name("account.schema.json").resolve().as_uri()}"
}}
"""
    )


def occupy_test():
    @config("subscription")
    class SubConf:
        organizations: List[str] = field(default_factory=list)
        """Organizations that you want to subscribe (tracks *all* repo events)"""

        repos: List[str] = field(default_factory=list)
        """Repositories that you want to subscribe
        Please avoid duplicating the organizations that you've subscribed.
        """

    with pytest.raises(NameError):

        @config("subscription")
        class SubConf2:
            organizations: List[str] = field(default_factory=list)
            """Organizations that you want to subscribe (tracks *all* repo events)"""

            repos: List[str] = field(default_factory=list)
            """Repositories that you want to subscribe
            Please avoid duplicating the organizations that you've subscribed.
            """


def repl_test():
    class SubConf:
        organizations: List[str] = field(default_factory=list)
        """Organizations that you want to subscribe (tracks *all* repo events)"""

    SubConf.__module__ = "_sre"  # Fake it as a non-retrievable class
    config("subscription")(SubConf)


def reload_test():

    sub_pth = Path("./temp/full_test/subscription.jsonc")
    sub_pth.write_text(
        """
{
    "organizations": ["GraiaProject", "GraiaCommunity"],
    "repos": ["GreyElaina/richuru"],
}
"""
    )

    # Initial creation
    @config("subscription")
    class SubConf:
        organizations: List[str] = field(default_factory=list)
        """Organizations that you want to subscribe (tracks *all* repo events)"""

    # Reload without instantiate
    @config("subscription")
    class SubConf:
        organizations: List[str] = field(default_factory=list)
        """Organizations that you want to subscribe (tracks *all* repo events)"""

    inst1 = create(SubConf)
    assert asdict(inst1) == {
        "organizations": ["GraiaProject", "GraiaCommunity"],
    }

    sub_pth.write_text(
        """
{
    "organizations": ["GraiaProject", "GraiaCommunity", "OctoOrg"],
    "repos": ["GreyElaina/richuru"],
}
"""
    )

    assert sys.getrefcount(inst1) == 3

    # Reload, without changes
    @config("subscription")
    class SubConf:
        organizations: List[str] = field(default_factory=list)
        """Organizations that you want to subscribe (tracks *all* repo events)"""

    inst2 = create(SubConf)
    assert asdict(inst2) == {
        "organizations": ["GraiaProject", "GraiaCommunity", "OctoOrg"],
    }

    assert inst1 is not inst2
    del inst1, inst2

    # Change schema
    @config("subscription")
    class SubConf:
        organizations: List[str] = field(default_factory=list)
        """Organizations that you want to subscribe (tracks *all* repo events)"""

        repos: List[str] = field(default_factory=list)
        """Repositories that you want to subscribe
        Please avoid duplicating the organizations that you've subscribed.
        """

    assert asdict(create(SubConf)) == {
        "organizations": ["GraiaProject", "GraiaCommunity", "OctoOrg"],
        "repos": ["GreyElaina/richuru"],
    }



from __future__ import annotations

import sys
from dataclasses import Field, dataclass, field
from dataclasses import fields as get_fields
from typing import TYPE_CHECKING, Callable, Tuple, Type, TypeVar, Union, cast

from loguru import logger
from typing_extensions import dataclass_transform

from .backend import dumps, loads
from .backend.types import JObject
from .schema_gen import ConfigModel
from .utils import from_dict, update

T = TypeVar("T")


@dataclass_transform(field_specifiers=(Field, field))
def config_stub(
    domain: str,
    *,
    init=True,
    repr=True,
    eq=True,
    order=False,
    unsafe_hash=False,
    frozen=False,
    match_args=True,
    kw_only=False,
    slots=False,
) -> Callable[[type[T]], type[T]]:
    """将类型转换为 kayaku 承认的设置类。

    `domain` 以外的参数将被直接传入 [`dataclasses.dataclass`][dataclasses.dataclass] 进行转换。

    Args:
        domain (str): *唯一的* 用于标识类的结构位置与类别的字符串，用 `.` 分开。

        init (bool, optional): If true (the default), a `__init__()` method will be generated.
        repr (bool, optional): If true (the default), a `__repr__()` method will be generated.
        eq (bool, optional): If true (the default), an `__eq__()` method will be generated.
        order (bool, optional): If specified, `__lt__()`, `__le__()`, `__gt__()`, and `__ge__()` methods will be generated.
        unsafe_hash (bool, optional): Force dataclass() to create a `__hash__()` method. Otherwise a `__hash__()` method is generated according to how eq and frozen are set.
        frozen (bool, optional): If specified, assigning to fields will generate an exception. This emulates read-only frozen instances.
        match_args (bool, optional): If true (the default is True), the `__match_args__` tuple will be created from the list of parameters to the generated `__init__()`.
        kw_only (bool, optional): If specified, then all fields will be marked as keyword-only.
        slots (bool, optional): If specified, `__slots__` attribute will be generated and new class will be returned instead of the original one.
    Returns:
        类的装饰器。
    """
    ...


def config_impl(domain: str, **kwargs) -> Callable[[type], Type[ConfigModel]]:
    def wrapper(cls: type) -> type[ConfigModel]:
        from . import domain as domain_mod

        if not hasattr(domain_mod, "_store"):
            raise RuntimeError("You cannot call `config` before initialization!")
        from .domain import _store as g_store
        from .domain import insert_domain

        cls = cast(Type[ConfigModel], dataclass(**kwargs)(cls))
        domain_tup: Tuple[str, ...] = tuple(domain.split("."))
        if not all(domain_tup):
            raise ValueError(f"{domain!r} contains empty segment!")

        if domain_tup in g_store.models:
            m_store = g_store.models[domain_tup]
            other = m_store.cls
            if (
                cls.__module__ == other.__module__
                and cls.__qualname__ == other.__qualname__
            ):
                if m_store.instance is not None:
                    instance = m_store.instance
                    if (
                        sys.getrefcount(instance) > 3
                    ):  # parameter, local var `instance`, `m_store.instance`
                        logger.warning(f"Instance of {other!r} is stilled referred!")
                    m_store.instance = None
                f_store = g_store.files[m_store.location.path]
                # remove original occupations

                g_store.cls_domains.pop(other)
                other_def_name = f_store.generator.retrieve_name(other)
                other_schema = f_store.generator.defs.pop(other_def_name)
                mount_dest = tuple(m_store.location.mount_dest)
                for field in get_fields(other):
                    sub_dest = mount_dest + (field.name,)
                    f_store.field_mount_record.remove(sub_dest)

                # inject new ones
                insert_domain(domain_tup, cls)
                cls_def_name = f_store.generator.retrieve_name(cls)
                cls_schema = f_store.generator.defs[cls_def_name]
                if cls_schema != other_schema:
                    logger.warning(f"Schema of {cls} has changed!")
                return cls

            raise NameError(f"{domain!r} is already occupied by {other!r}")
        insert_domain(domain_tup, cls)
        return cls

    return wrapper


config = config_stub if TYPE_CHECKING else config_impl


def create(cls: Type[T], flush: bool = False) -> T:
    """创建对应的模型。

    给 `flush` 传入真值会强制重载。

    Note:
        为了安全重载，我们推荐你让返回的实例存活时间尽量短。
    """

    from .domain import _store

    if not (issubclass(cls, ConfigModel) and cls in _store.cls_domains):
        raise TypeError(f"{cls!r} is not a registered ConfigModel class!")
    domain = _store.cls_domains[cls]
    model_store = _store.models[domain]
    if flush or model_store.instance is None:

        fmt_path = model_store.location
        document = loads(fmt_path.path.read_text("utf-8") or "{}")
        container = document
        for sect in fmt_path.mount_dest:
            container = container.setdefault(sect, JObject())
        model_store.instance = from_dict(cls, container)

    return cast(T, model_store.instance)


def save(model: Union[T, Type[T]]) -> None:
    """保存某个类的设置。 对应的 JSON Schema 也会被更新。"""
    from .domain import _store

    cls = cast(Type[ConfigModel], model if isinstance(model, type) else model.__class__)
    m_store = _store.models[_store.cls_domains[cls]]
    inst = m_store.instance
    if inst is not None:
        document = loads(m_store.location.path.read_text("utf-8") or "{}")
        container = document
        for sect in m_store.location.mount_dest:
            container = container.setdefault(sect, JObject())
        update(container, inst)
        document.pop("$schema", None)
        document["$schema"] = m_store.location.path.with_suffix(".schema.json").as_uri()
        m_store.location.path.write_text(
            dumps(_store.prettifier.prettify(document), endline=True), "utf-8"
        )
    m_store.location.path.with_suffix(".schema.json").write_text(
        dumps(_store.files[m_store.location.path].get_schema()), "utf-8"
    )



from pathlib import Path

import pytest

from kayaku.spec import (
    FormattedPath,
    PathFill,
    PathSpec,
    SectionSpec,
    SourceSpec,
    parse_path,
    parse_source,
)


def test_spec_parse():
    assert parse_source("{module.**}.secrets") == SourceSpec(
        ["module"], ["secrets"], SectionSpec(["module"], [])
    )

    assert parse_source("graia.{**}") == SourceSpec(["graia"], [], SectionSpec([], []))
    assert parse_path("./config/modules/{}::config.{**}") == PathSpec(
        [".", "config", "modules", PathFill.SINGLE], ["config", PathFill.EXTEND]
    )


def test_parse_spec_err():
    with pytest.raises(ValueError):
        parse_source("*")
    with pytest.raises(ValueError):
        parse_path("{**}::{**}")


def test_format_path_spec():
    base_pth = Path("./temp/spec/").resolve()
    base_pth.mkdir(parents=True, exist_ok=True)
    assert parse_path(base_pth.as_posix() + "/{}::config.{**}.{}.mock").format(
        ["a", "b", "c", "d"]
    ) == FormattedPath(Path(base_pth / "a"), ["config", "b", "c", "d", "mock"])

    assert parse_path(base_pth.as_posix() + "/{}::config.{}.mock").format(
        ["a", "b"]
    ) == FormattedPath(Path(base_pth / "a"), ["config", "b", "mock"])

    assert parse_path(base_pth.as_posix() + "/{**}::config.{}.{}.mock").format(
        ["a", "b", "c", "d"]
    ) == FormattedPath(Path(base_pth / "a/b"), ["config", "c", "d", "mock"])

    assert (
        parse_path(base_pth.as_posix() + "/{}:config.{}.mock").format(
            ["a", "b", "c", "d"]
        )
        is None
    )
    assert parse_path(base_pth.as_posix() + "/{**}").format(
        ["a", "b", "c"]
    ) == FormattedPath(Path(base_pth / "a/b/c"), [])



"""dataclass schema generator

Modified from https://github.com/Peter554/dc_schema.


MIT License

Copyright (c) 2022 Peter Byfield

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from __future__ import annotations

import dataclasses
import datetime
import enum
import inspect
import numbers
import re
import sys
import types
import typing as t
from abc import ABC

import typing_extensions as t_e

from kayaku.doc_parse import store_field_description

_MISSING = dataclasses.MISSING


_Format = t.Literal[
    "date-time",
    "time",
    "date",
    "duration",
    "email",
    "idn-email",
    "hostname",
    "idn-hostname",
    "ipv4",
    "ipv6",
    "uuid",
    "uri",
    "uri-reference",
    "iri",
    "iri-reference",
    "regex",
]

SCHEMA_ANNO_KEY_MAP = {
    "min_length": "minLength",
    "max_length": "maxLength",
    "exclusive_minimum": "exclusiveMinimum",
    "exclusive_maximum": "exclusiveMaximum",
    "multiple_of": "multipleOf",
    "min_items": "minItems",
    "max_items": "maxItems",
    "unique_items": "uniqueItems",
}


@dataclasses.dataclass(frozen=True)
class Schema:
    title: t.Optional[str] = None
    description: t.Optional[str] = None
    examples: t.Optional[list[t.Any]] = None
    deprecated: t.Optional[bool] = None

    def schema(self):
        return {
            SCHEMA_ANNO_KEY_MAP.get(k, k): v
            for k, v in dataclasses.asdict(self).items()
            if v is not None
        }


@dataclasses.dataclass(frozen=True)
class StringSchema(Schema):
    min_length: t.Optional[int] = None
    max_length: t.Optional[int] = None
    pattern: t.Optional[str] = None
    format: t.Optional[_Format] = None


@dataclasses.dataclass(frozen=True)
class NumberSchema(Schema):
    minimum: t.Optional[numbers.Number] = None
    maximum: t.Optional[numbers.Number] = None
    exclusive_minimum: t.Optional[numbers.Number] = None
    exclusive_maximum: t.Optional[numbers.Number] = None
    multiple_of: t.Optional[numbers.Number] = None


@dataclasses.dataclass(frozen=True)
class ContainerSchema(Schema):
    min_items: t.Optional[int] = None
    max_items: t.Optional[int] = None
    unique_items: t.Optional[bool] = None


class ConfigModel(ABC):
    __dataclass_fields__: t.ClassVar[t.Dict[str, dataclasses.Field]]

    @classmethod
    def __subclasshook__(cls, oth):
        return isinstance(oth, type) and dataclasses.is_dataclass(oth)


def is_sub_type(sub: t.Any, parent: t.Any) -> bool:
    sub_origin = t_e.get_origin(sub) or sub
    return (
        isinstance(sub_origin, type)
        and issubclass(sub_origin, parent)
        or sub_origin == parent
    )


Unions = (t.Union,) if sys.version_info < (3, 10) else (t.Union, types.UnionType)


class SchemaGenerator:
    def __init__(self, dc: t.Type[ConfigModel] | None = None) -> None:
        self.root = dc
        self.seen_root = False
        self.defs = {}

    def retrieve_name(self, typ: t.Type) -> str:
        return f"{typ.__module__}.{typ.__qualname__}"

    def retrieve_title(self, typ: t.Type) -> str:
        return self.retrieve_name(typ)

    def format_docstring_description(
        self, field: dataclasses.Field, description: str
    ) -> str | None:
        return description

    @classmethod
    def from_dc(cls, dc: t.Type[ConfigModel]) -> dict[str, t.Any]:
        generator = cls(dc)
        schema = generator.get_dc_schema(dc)
        if generator.defs:
            schema["$defs"] = generator.defs

        return {
            "$schema": "https://json-schema.org/draft/2020-12/schema",
            **schema,
        }

    def get_dc_schema(self, dc: t.Type[ConfigModel]) -> dict[str, t.Any]:
        if dc == self.root:
            if self.seen_root:
                return {"$ref": "#"}
            self.seen_root = True
            return self.create_dc_schema(dc)
        else:
            name = self.retrieve_name(dc)
            if name not in self.defs:
                schema = self.create_dc_schema(dc)
                self.defs[name] = schema
            return {
                "$ref": f"#/$defs/{name}",
            }

    def create_dc_schema(self, dc: t.Type[ConfigModel]):
        schema = {
            "type": "object",
            "title": self.retrieve_title(dc),
            "properties": {},
            "required": [],
        }
        store_field_description(dc, dc.__dataclass_fields__)
        type_hints = t_e.get_type_hints(dc, include_extras=True)
        if (
            dc.__doc__ is not None
            and dc.__doc__
            != f"{dc.__name__}{str(inspect.signature(dc)).replace(' -> None', '')}"  # Ignore the generated __doc__
        ):
            schema["description"] = dc.__doc__
        for field in dataclasses.fields(dc):
            typ = type_hints[field.name]
            if (f_description := field.metadata.get("description")) is not None:
                f_a: Schema
                base, f_a = (
                    t_e.get_args(typ)
                    if t_e.get_origin(typ) == t_e.Annotated
                    else (typ, Schema())
                )
                if f_a.description is None:
                    object.__setattr__(
                        f_a,
                        "description",
                        self.format_docstring_description(field, f_description),
                    )
                typ = t_e.Annotated[base, f_a]
            schema["properties"][field.name] = self.get_field_schema(typ, field.default)
            field_is_optional = (
                field.default is not _MISSING or field.default_factory is not _MISSING
            )
            if not field_is_optional:
                schema["required"].append(field.name)
        if not schema["required"]:
            schema.pop("required")
        return schema

    def get_simple_schema(self, typ: t.Type, default: t.Any):
        if typ in (t.Any, t_e.Any):
            return self.get_any_schema(default)
        elif typ in (None, type(None)):
            return self.get_none_schema(default)
        elif is_sub_type(typ, str):
            return self.get_str_schema(default)
        elif is_sub_type(typ, bool):
            return self.get_bool_schema(default)
        elif is_sub_type(typ, int):
            return self.get_int_schema(default)
        elif is_sub_type(typ, numbers.Number):
            return self.get_number_schema(default)
        elif is_sub_type(typ, enum.Enum):
            return self.get_enum_schema(typ, default)
        elif is_sub_type(typ, datetime.datetime):
            return self.get_datetime_schema()
        elif is_sub_type(typ, datetime.date):
            return self.get_date_schema()
        elif is_sub_type(typ, re.Pattern):
            return self.get_regex_schema()

    def get_complex_schema(self, typ: t.Type, default: t.Any):
        if dataclasses.is_dataclass(typ):
            return self.get_dc_schema(typ)
        elif t_e.get_origin(typ) in Unions:
            return self.get_union_schema(typ, default)
        elif t_e.get_origin(typ) == t.Literal:
            return self.get_literal_schema(typ, default)
        elif t_e.get_origin(typ) == t_e.Annotated:
            return self.get_annotated_schema(typ, default)
        elif t_e.is_typeddict(typ):
            return self.get_typed_dict_schema(typ)
        elif is_sub_type(typ, dict):
            return self.get_dict_schema(typ)
        elif is_sub_type(typ, list):
            return self.get_list_schema(typ)
        elif is_sub_type(typ, tuple):
            return self.get_tuple_schema(typ, default)
        elif is_sub_type(typ, set):
            return self.get_set_schema(typ)

    def get_field_schema(self, typ: t.Type, default: t.Any):
        if (schema := self.get_simple_schema(typ, default)) is not None:
            return schema
        if (schema := self.get_complex_schema(typ, default)) is not None:
            return schema
        raise NotImplementedError(f"field type '{typ}' not implemented")

    def get_any_schema(self, default: t.Any):
        return {} if default is _MISSING else {"default": default}

    def get_union_schema(self, typ: t.Type, default: t.Any):
        args = t_e.get_args(typ)
        if default is _MISSING:
            return {
                "anyOf": [self.get_field_schema(arg, _MISSING) for arg in args],
            }
        else:
            return {
                "anyOf": [self.get_field_schema(arg, _MISSING) for arg in args],
                "default": default,
            }

    def get_literal_schema(self, typ, default):
        schema = {} if default is _MISSING else {"default": default}
        args = t_e.get_args(typ)
        return {"enum": list(args), **schema}

    def get_dict_schema(self, typ):
        args = t_e.get_args(typ)
        assert len(args) in {0, 2}
        if args:
            assert args[0] == str
            return {
                "type": "object",
                "additionalProperties": self.get_field_schema(args[1], _MISSING),
            }
        else:
            return {"type": "object"}

    def get_typed_dict_schema(self, typ: type[t.TypedDict]):
        fields: list[tuple[str, t.Any]] = []
        required: list[str] = []  # Python 3.8- don't have `__required_keys__`
        for name, anno in t_e.get_type_hints(typ, include_extras=True).items():
            anno: t.Any
            schema_anno: Schema | None = None
            if t_e.get_origin(anno) == t_e.Annotated:
                anno, schema_anno = t_e.get_args(anno)
            if t_e.get_origin(anno) == t_e.Required:
                required.append(name)
            elif typ.__total__ and t_e.get_origin(anno) != t_e.NotRequired:
                required.append(name)
            if t_e.get_origin(anno) in (t_e.Required, t_e.NotRequired):
                anno = t_e.get_args(anno)[0]
            fields.append(
                (
                    name,
                    (anno if schema_anno is None else t_e.Annotated[anno, schema_anno]),
                )
            )
        dc: type[ConfigModel] = t.cast(
            t.Type[ConfigModel], dataclasses.make_dataclass(typ.__qualname__, fields)
        )
        dc.__module__ = typ.__module__
        store_field_description(typ, dc.__dataclass_fields__)
        dc_schema = self.get_dc_schema(dc)
        dc_def = self.defs[self.retrieve_name(dc)]
        dc_def["required"] = required
        return dc_schema

    def get_list_schema(self, typ):
        args = t_e.get_args(typ)
        assert len(args) in {0, 1}
        if args:
            return {
                "type": "array",
                "items": self.get_field_schema(args[0], _MISSING),
            }
        else:
            return {"type": "array"}

    def get_tuple_schema(self, typ, default):
        schema = {} if default is _MISSING else {"default": list(default)}
        args = t_e.get_args(typ)
        if args and len(args) == 2 and args[1] is ...:
            schema = {
                "type": "array",
                "items": self.get_field_schema(args[0], _MISSING),
                **schema,
            }
        elif args:
            schema = {
                "type": "array",
                "prefixItems": [self.get_field_schema(arg, _MISSING) for arg in args],
                "minItems": len(args),
                "maxItems": len(args),
                **schema,
            }
        else:
            schema = {"type": "array", **schema}
        return schema

    def get_set_schema(self, typ):
        args = t_e.get_args(typ)
        assert len(args) in {0, 1}
        if args:
            return {
                "type": "array",
                "items": self.get_field_schema(args[0], _MISSING),
                "uniqueItems": True,
            }
        else:
            return {"type": "array", "uniqueItems": True}

    def get_none_schema(self, default):
        if default is _MISSING:
            return {"type": "null"}
        else:
            return {"type": "null", "default": default}

    def get_str_schema(self, default):
        if default is _MISSING:
            return {"type": "string"}
        else:
            return {"type": "string", "default": default}

    def get_bool_schema(self, default):
        if default is _MISSING:
            return {"type": "boolean"}
        else:
            return {"type": "boolean", "default": default}

    def get_int_schema(self, default):
        if default is _MISSING:
            return {"type": "integer"}
        else:
            return {"type": "integer", "default": default}

    def get_number_schema(self, default):

        if default is _MISSING:
            return {"type": "number"}
        else:
            return {"type": "number", "default": default}

    def get_enum_schema(self, typ: type[enum.Enum], default: enum.Enum):
        name = self.retrieve_name(typ)
        title = self.retrieve_title(typ)
        if name not in self.defs:
            self.defs[name] = {
                "title": title,
                "enum": [v.value for v in typ],
            }
        if default is _MISSING:
            return {
                "$ref": f"#/$defs/{name}",
            }
        else:
            return {
                "$ref": f"#/$defs/{name}",
                "default": default.value,
            }

    def get_annotated_schema(self, typ, default):
        args = t_e.get_args(typ)
        assert len(args) == 2
        base, annotation = args
        assert isinstance(annotation, Schema)
        if isinstance(annotation, StringSchema):
            if not is_sub_type(base, str):
                raise TypeError(f"Trying to apply string-specific annotation to {base}")
        elif isinstance(annotation, ContainerSchema):
            if not any(is_sub_type(base, typ) for typ in (list, tuple, set)):
                raise TypeError(
                    f"Trying to apply sequence-specific annotation to {base}"
                )
        elif isinstance(annotation, NumberSchema):
            if not (is_sub_type(base, numbers.Number) and base is not bool):
                raise TypeError(f"Trying to apply number-specific annotation to {base}")
        schema = self.get_field_schema(base, default)
        schema.update(annotation.schema())
        return schema

    def get_datetime_schema(self):
        return {"type": "string", "format": "date-time"}

    def get_date_schema(self):
        return {"type": "string", "format": "date"}

    def get_regex_schema(self):
        return {"type": "string", "format": "regex"}


def write_schema_ref(root: dict, sect: tuple[str, ...], name: str) -> None:
    for s in sect:
        root["type"] = "object"
        required: list[str] = root.setdefault("required", [])
        if s not in required:
            required.append(s)
        root = root.setdefault("properties", {})
        root = root.setdefault(s, {})

    all_of: list[dict[str, str]] = root.setdefault("allOf", [])
    if {"$ref": f"#/$defs/{name}"} not in all_of:
        all_of.append({"$ref": f"#/$defs/{name}"})



from __future__ import annotations

import ast
import re
from typing import Any, cast

from lark.lexer import Token
from lark.visitors import Transformer as BaseTransformer
from lark.visitors import merge_transformers, v_args

from . import wsc
from .types import (
    WSC,
    Array,
    Float,
    HexInteger,
    Identifier,
    Integer,
    JContainer,
    JNumber,
    JObject,
    JString,
    JType,
    JWrapper,
    Key,
    Member,
    Quote,
    Value,
)


class Transformer(BaseTransformer):
    """
    A [Transformer][lark.visitors.Transformer] for JSON5
    """

    @v_args(inline=True)
    def string(self, string: JString):
        return string

    @v_args(inline=True)
    def SINGLE_QUOTE_CHARS(self, token: Token) -> tuple[str, list[int]]:
        return ast.literal_eval(f"'{token.value}'"), [
            m.start() for m in re.finditer("\\\n", token.value)
        ]

    @v_args(inline=True)
    def DOUBLE_QUOTE_CHARS(self, token: Token) -> tuple[str, list[int]]:
        return ast.literal_eval(f'"{token.value}"'), [
            m.start() for m in re.finditer("\\\n", token.value)
        ]

    @v_args(inline=True)
    def double_quote_string(
        self, string: tuple[str, list[int]] | None = None
    ) -> JString:
        value, linebreaks = string or ("", [])
        s = JString(value)
        s.__post_init__(quote=Quote.DOUBLE, linebreaks=linebreaks)
        return s

    @v_args(inline=True)
    def single_quote_string(
        self, string: tuple[str, list[int]] | None = None
    ) -> JString:
        value, linebreaks = string or ("", [])
        s = JString(value)
        s.__post_init__(quote=Quote.SINGLE, linebreaks=linebreaks)
        return s

    @v_args(inline=True)
    def IDENTIFIER_NAME(self, string) -> Identifier:
        return Identifier(string)

    @v_args(inline=True)
    def number(self, number: JNumber):
        return number

    def SIGNED_HEXNUMBER(self, token: Token):
        i = HexInteger(int(token.value, base=16))
        i.__post_init__(token.value)
        return i

    def SIGNED_NUMBER(self, token: Token):
        if (
            "." not in token.value
            and "e" not in token.value
            and token.value
            not in {"NaN", "+NaN", "-NaN", "+Infinity", "-Infinity", "Infinity"}
        ):
            i = Integer(token.value)
            i.__post_init__(token.value)
            return i
        f = Float(token.value)
        f.__post_init__(token.value)
        return f

    @staticmethod
    def _set_trail(obj: JContainer, children: list) -> None:
        obj.json_container_tail = children[-2]
        obj.json_container_trailing_comma = (
            isinstance(children[-3], Token) and children[-3].value == ","
        )

    def object(self, children: list) -> Any:
        o = JObject(cast(Member, c) for c in children if isinstance(c, tuple))
        self._set_trail(o, children)
        return o

    def array(self, children: list) -> Any:
        a = Array(
            (cast(Value, value) for value in children if isinstance(value, JType))
        )
        self._set_trail(a, children)
        return a

    @v_args(inline=True)
    def literal(self, token: Token):
        if token.value == "true":
            return JWrapper[bool](True)
        elif token.value == "false":
            return JWrapper[bool](False)
        assert token.value == "null"
        return JWrapper[None](None)

    @v_args(inline=True)
    def pack_wsc(self, before: list[WSC], value: JType, after: list[WSC]) -> JType:
        value.json_before = before
        value.json_after = after
        return value

    def member(self, kv: list[Key | Value]) -> Member:
        assert len(kv) == 2
        return (cast(Key, kv[0]), cast(Value, kv[1]))

    value = pack_wsc
    key = pack_wsc

    pair = tuple


transformer = merge_transformers(Transformer(), wsc=wsc.transformer)



import inspect

from kayaku import backend as json5
from kayaku.backend.types import Quote, convert
from kayaku.pretty import Prettifier


def test_pretty_single_wrapped():
    origins = [convert(i) for i in [{"a": "b"}, {"a": 1}, [1], ["acc"], [], {}]]
    expected = ["""{"a": "b"}""", """{"a": 1}""", """[1]""", """["acc"]""", "[]", "{}"]

    for o, e in zip(origins, expected):
        assert json5.dumps(Prettifier().prettify(o)) == e


def test_pretty_single_unwrapped():
    origin = convert({"a": "b"})
    expected_unwrapped = inspect.cleandoc(
        """\
        {
            "a": "b"
        }
        """
    )
    assert (
        json5.dumps(Prettifier(unfold_single=True).prettify(origin))
        == expected_unwrapped
    )
    origin = convert({"a": 1})
    expected_unwrapped = inspect.cleandoc(
        """\
        {
            "a": 1
        }
        """
    )
    assert (
        json5.dumps(Prettifier(unfold_single=True).prettify(origin))
        == expected_unwrapped
    )


def test_pretty_complex():
    obj = {"a": 5, "b": 6, "c": [None, {"a": "b", "c": {"d": "e"}}, "F"]}
    result = inspect.cleandoc(
        """\
        {
            "a": 5,
            "b": 6,
            "c": [
                null,
                {
                    "a": "b",
                    "c": {"d": "e"}
                },
                "F"
            ]
        }
        """
    )
    assert json5.dumps(Prettifier().prettify(convert(obj))) == result


def test_pretty_flavor():
    input_str = "{a: 6}"
    assert (
        json5.dumps(Prettifier(key_quote=None).prettify(json5.loads(input_str)))
        == "{a: 6}"
    )
    assert (
        json5.dumps(Prettifier(key_quote=Quote.DOUBLE).prettify(json5.loads(input_str)))
        == '{"a": 6}'
    )
    assert (
        json5.dumps(Prettifier(key_quote=Quote.SINGLE).prettify(json5.loads(input_str)))
        == "{'a': 6}"
    )
    assert (
        json5.dumps(Prettifier(key_quote=False).prettify(json5.loads("{'a': 6}")))
        == "{a: 6}"
    )
    assert (
        json5.dumps(Prettifier(key_quote=False).prettify(json5.loads("{'.': 6}")))
        == '{".": 6}'
    )


def test_pretty_comment():
    input_str = """
    {
        "a": 5,
        "b": 6,
        "c": [
            null,
            /*massive*/{ //plots
                "a": "b",
                "c": {"d": "e"}
            },
            "F", // marvelous
        ],
        "ariadne":{
            /*
            More annotations


            @type: int
             */
            "account": 0
        },
        "broadcast":{
            /*@type: List[dict]*/
            "dct":[
                {"3": 5} // what a mess
            ] /*input*/,
        /*
                 * Test annotating
                 * contain '*' and '/'
             * shorter ident
                         * longer ident
                 *
                 * @type: Optional[P]
                                */
            "p": null
        }, // bro
    }
    """

    output = inspect.cleandoc(
        """\
        {
            "a": 5,
            "b": 6,
            "c": [
                null,
                /*massive*/
                { //plots
                    "a": "b",
                    "c": {"d": "e"}
                },
                "F" // marvelous
            ],
            "ariadne": {
                /*
                 * More annotations
                 *
                 *
                 * @type: int
                 */
                "account": 0
            },
            "broadcast": {
                /*@type: List[dict]*/
                /*input*/
                "dct": [
                    {"3": 5} // what a mess
                ],
                /*
                 * Test annotating
                 * contain '*' and '/'
                 * shorter ident
                 * longer ident
                 *
                 * @type: Optional[P]
                 */
                "p": null
            } // bro
        }
        """
    )

    assert json5.dumps(Prettifier().prettify(json5.loads(input_str))) == output



from __future__ import annotations

from typing import Any, Callable, TextIO

from . import wsc
from .types import AnyNumber, Float, Identifier, Integer, JNumber, JString, JWrapper

ESCAPES = {
    "\\": r"\\",
    "\n": r"\n",
    "\r": r"\r",
    "\b": r"\b",
    "\f": r"\f",
    "\t": r"\t",
    "\v": r"\v",
    "\0": r"\0",
    "\u2028": r"\\u2028",
    "\u2029": r"\\u2029",
}

QUOTE_ESCAPE = {"'": {"'": r"\'"}, '"': {'"': r"\""}}

JSONEncoderMethod = Callable[[Any, Any], None]
"""A JSON encoder type method"""


def with_style(fn: JSONEncoderMethod) -> JSONEncoderMethod:
    """
    A decorator providing whitespaces and comments handling for encoders.

    It handles `json_before` and `json_after` serialization
    if the object to encode has these attributes.

    :param fn: An encoder method for a spcific type.
    """

    def encode_with_style(self: Encoder, obj: Any) -> None:
        self.fp.write(
            "".join(wsc.encode_wsc(w) for w in getattr(obj, "json_before", [])),
        )
        fn(self, obj)
        self.fp.write(
            "".join(wsc.encode_wsc(w) for w in getattr(obj, "json_after", [])),
        )

    return encode_with_style


def escape_string(string: str, **escapes: str | int | None) -> str:
    out = string.translate(str.maketrans({**escapes, **ESCAPES}))
    if isinstance(string, JString):
        for line_break in string.linebreaks:
            out = out[: line_break - 1] + "\\\n" + out[line_break - 1 :]
    return out


class Encoder:
    def __init__(self, fp: TextIO):
        self.encode_func: dict[tuple[type, ...], Callable[[Any], None]] = {
            (JWrapper,): self.encode_wrapper,
            (bool, type(None)): self.encode_bool_like,
            (JNumber, int, float): self.encode_number,
            (str,): self.encode_string,
            (dict,): self.encode_dict,
            (list, tuple): self.encode_iterable,
        }
        self.fp = fp

    def encode(self, obj: Any) -> None:
        for typ_tuple, func in self.encode_func.items():
            if isinstance(obj, typ_tuple):
                func(obj)
                return
        raise NotImplementedError(f"Unknown type: {type(obj)}")

    @with_style
    def encode_wrapper(self, obj: JWrapper) -> None:
        return self.encode(obj.value)

    def encode_bool_like(self, obj: bool | None) -> None:
        self.fp.write({True: "true", False: "false", None: "null"}[obj])

    @with_style
    def encode_dict(self, obj: dict) -> None:
        self.fp.write("{")
        self.fp.write(
            "".join(wsc.encode_wsc(w) for w in getattr(obj, "json_container_head", []))
        )
        if obj:
            last_k = next(reversed(obj))
            for k, v in obj.items():
                self.encode_pair(k, v)
                if k is not last_k or getattr(
                    obj, "json_container_trailing_comma", False
                ):
                    self.fp.write(",")
        self.fp.write(
            "".join(wsc.encode_wsc(w) for w in getattr(obj, "json_container_tail", []))
        )
        self.fp.write("}")

    @with_style
    def encode_iterable(self, obj: list | tuple) -> None:
        self.fp.write("[")
        self.fp.write(
            "".join(wsc.encode_wsc(w) for w in getattr(obj, "json_container_head", []))
        )
        if obj:
            last_v = obj[-1]
            for v in obj:
                self.encode(v)
                if v is not last_v or getattr(
                    obj, "json_container_trailing_comma", False
                ):
                    self.fp.write(",")
        self.fp.write(
            "".join(wsc.encode_wsc(w) for w in getattr(obj, "json_container_tail", []))
        )
        self.fp.write("]")

    def encode_pair(self, key: str, value: Any) -> None:
        self.encode(key)
        self.fp.write(":")
        self.encode(value)

    @with_style
    def encode_number(self, obj: AnyNumber | int | float) -> None:
        if not isinstance(obj, JNumber):
            obj = (Integer if isinstance(obj, int) else Float)(obj)
        self.fp.write(obj.__round_dump__())

    @with_style
    def encode_string(self, obj: str) -> None:
        if isinstance(obj, Identifier):
            self.fp.write(obj)
        else:
            quote = obj.quote.value if isinstance(obj, JString) else '"'
            self.fp.write(f"{quote}{escape_string(obj, **QUOTE_ESCAPE[quote])}{quote}")



from .domain import bootstrap as bootstrap
from .domain import initialize as initialize
from .domain import save_all as save_all
from .model import config as config
from .model import create as create
from .model import save as save
from .utils import copying_field

default = copying_field



from __future__ import annotations

import enum
import re
from contextlib import suppress
from contextvars import ContextVar
from dataclasses import dataclass
from pathlib import Path
from typing import Literal, cast

from typing_extensions import TypedDict

_LIG = r"[A-Za-z0-9_-]"
_L_REP = rf"({_LIG}+\.)*"
_R_REP = rf"(\.{_LIG}+)*"
SOURCE_REGEX = re.compile(
    rf"(?P<prefix>{_L_REP}){{(?P<sect_prefix>{_L_REP})\*\*(?P<sect_suffix>{_R_REP})}}(?P<suffix>{_R_REP})"
)


@dataclass
class SectionSpec:
    prefix: list[str]
    suffix: list[str]


@dataclass
class SourceSpec:
    prefix: list[str]
    suffix: list[str]
    section: SectionSpec


class SourceRegexGroup(TypedDict):
    prefix: str
    sect_prefix: str
    sect_suffix: str
    suffix: str


def parse_source(spec: str) -> SourceSpec:
    if not (match_res := SOURCE_REGEX.fullmatch(spec)):
        raise ValueError(f"{spec!r} doesn't match {SOURCE_REGEX.pattern!r}")
    groups: SourceRegexGroup = cast(SourceRegexGroup, match_res.groupdict())
    section: SectionSpec = SectionSpec(
        groups["sect_prefix"].split(".")[:-1], groups["sect_suffix"].split(".")[1:]
    )
    return SourceSpec(
        groups["prefix"].split(".")[:-1] + section.prefix,
        section.suffix + groups["suffix"].split(".")[1:],
        section,
    )


class PathFill(enum.Enum):
    SINGLE = object()  # {} or {*}
    EXTEND = object()  # {**}


@dataclass
class PathSpec:
    path: list[str | PathFill]
    section: list[str | PathFill]

    @property
    def fill_lens(self) -> tuple[int, int]:
        fills = [i for i in self.path + self.section if isinstance(i, PathFill)]
        if PathFill.EXTEND not in fills:
            return len(fills), 0
        ext_ind = fills.index(PathFill.EXTEND)
        return ext_ind, len(fills) - ext_ind - 1

    def format(self, parts: list[str]) -> FormattedPath | None:
        front_len, back_len = self.fill_lens
        front = parts[:front_len]
        back = parts[front_len:][-back_len or len(parts) :]
        ext = parts[front_len : -back_len or len(parts)]
        if (
            len(front) != front_len
            or len(back) != back_len
            or (ext and PathFill.EXTEND not in self.path + self.section)
        ):
            return
        formatted_it = iter(front + back)

        fmt_path: list[str] = []
        for p in self.path:
            if p is PathFill.EXTEND:
                fmt_path.extend(ext)
            else:
                fmt_path.append(next(formatted_it) if p is PathFill.SINGLE else p)
        fmt_sect: list[str] = []
        for p in self.section:
            if p is PathFill.EXTEND:
                fmt_sect.extend(ext)
            else:
                fmt_sect.append(next(formatted_it) if p is PathFill.SINGLE else p)
        return FormattedPath(Path("/".join(fmt_path)), fmt_sect)  # Allow absolute path


@dataclass
class FormattedPath:
    path: Path
    mount_dest: list[str]

    def __post_init__(self) -> None:
        if not self.path.suffix:
            self.path = self.path.with_suffix(".jsonc")  # TODO: Config
        self.path.parent.mkdir(parents=True, exist_ok=True)
        self.path.touch(exist_ok=True)
        self.path = self.path.resolve()


def parse_path(spec: str) -> PathSpec:
    replacer = {"{*}": PathFill.SINGLE, "{}": PathFill.SINGLE, "{**}": PathFill.EXTEND}
    location, section = spec.rsplit("::", 1) if "::" in spec else (spec, "")
    path_parts: list[str | PathFill] = [replacer.get(l, l) for l in location.split("/")]
    section_parts: list[str | PathFill] = (
        [replacer.get(l, l) for l in section.split(".")] if section else []
    )
    if path_parts.count(PathFill.EXTEND) + section_parts.count(PathFill.EXTEND) > 1:
        raise ValueError(f"""Found more than one "extend" part ({{**}}) in {spec}""")
    return PathSpec(path_parts, section_parts)



"""
This modules contains all shared and reusable types supporting style preservation.

Most of the time you won't have to instantiate them manually.
"""

from __future__ import annotations

import enum
import math
import re
from datetime import date, datetime, time
from enum import Enum
from typing import Any, Generic, List, Tuple, TypeVar, overload

from typing_extensions import Self, TypeAlias


class JType:
    """
    Base class for parsed types with style and metadata preservation.
    """

    json_before: list[WSC] = []
    """Whitespaces and comments sequence before the object."""
    json_after: list[WSC] = []
    """Whitespaces and comments sequence after the object."""

    def __post_init__(self) -> Self:
        self.json_before = []
        self.json_after = []
        return self

    def __repr__(self) -> str:
        if attrs := getattr(self, "__dict__"):
            kwargs = ", ".join(f"{k}={v}" for k, v in attrs.items())
            return f"{self.__class__.__name__}({super().__repr__()}, {kwargs})"
        return f"{self.__class__.__name__}({super().__repr__()})"

    def __json_clear__(self):
        self.json_before = []
        self.json_after = []


class JContainer(JType):
    """
    Base class for containers with style and metadata preservation.
    """

    json_container_tail: list[WSC] = []
    """Whitespaces and comments sequence in the tail of the container."""
    json_container_trailing_comma: bool = False
    """Wether this container have a trailing comma or not."""

    def __post_init__(
        self,
        tail: list[WSC] | None = None,
        trailing_comma: bool = False,
    ) -> Self:
        super().__post_init__()

        self.json_container_tail = tail or []
        self.json_container_trailing_comma = trailing_comma
        return self


class WhiteSpace(str):
    """Stores a sequence of whitespaces"""

    def __repr__(self) -> str:
        return f"WhiteSpace({super().__repr__()})"


class Comment(str):
    """Store a comment"""

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({super().__repr__()})"


WSC: TypeAlias = "WhiteSpace | Comment"
"""A whitespace or a comment"""


class BlockStyleComment(Comment):
    """Stores a block-style comment (ie. starting with `/*` and ending with `*/`)"""


class LineStyleComment(Comment):
    """Stores a line-style comment (ie. starting with `//` and ending at the end of the current line)"""


class JObject(JContainer, dict):
    """A JSON Object with order and style preservation"""


T = TypeVar("T")


class Array(JContainer, List[T]):
    """A JSON Array with style preservation"""


class Identifier(JType, str):
    "A quoteless string without special characters"

    def __repr__(self) -> str:
        if attrs := getattr(self, "__dict__"):
            kwargs = ", ".join(f"{k}={v}" for k, v in attrs.items())
            return f"{self.__class__.__name__}({super().__repr__()}, {kwargs})"
        return f"{self.__class__.__name__}({super().__repr__()})"


Ident = Identifier


class Quote(Enum):
    """Known quotes formats"""

    value: str
    SINGLE = "'"
    DOUBLE = '"'


class JString(JType, str):
    """A JSON String with style preservation"""

    quote: Quote = Quote.DOUBLE
    """Quote character wrapping the string"""

    linebreaks: list[int] = []
    """Escaped line breaks positions"""

    def __post_init__(
        self,
        quote: str | Quote = Quote.DOUBLE,
        linebreaks: list[int] | None = None,
    ) -> Self:
        super().__post_init__()
        self.quote = Quote(quote) if isinstance(quote, str) else quote
        self.linebreaks = linebreaks or []
        return self

    def __repr__(self) -> str:
        if attrs := getattr(self, "__dict__"):
            kwargs = ", ".join(f"{k}={v}" for k, v in attrs.items())
            return f"{self.__class__.__name__}({self.quote.value}{self}{self.quote.value}, {kwargs})"
        return f"{self.__class__.__name__}({super().__repr__()})"


Key: TypeAlias = "JString | Identifier | str"


class JNumber(JType):
    """
    Base class for all Number types and representations.
    """

    origin: str | None = None

    def __post_init__(self, origin: str | None = None) -> Self:
        self.origin = origin
        return super().__post_init__()

    def __round_dump__(self) -> str:
        ...


class Integer(JNumber, int):
    """
    A JSON integer compatible with Python's `int`.
    """

    def __str__(self) -> str:
        return int.__repr__(self)

    def __round_dump__(self) -> str:
        if self.origin and int(self.origin) == self:
            return self.origin

        return int.__repr__(self)


class HexInteger(Integer):
    """
    A JSON integer compatible with Python's `int` and represented in its hexadecimal form.
    """

    def __str__(self) -> str:
        return hex(self)

    def __round_dump__(self) -> str:
        if self.origin and int(self.origin, base=16) == self:
            return self.origin

        return hex(self)


class Float(JNumber, float):
    """
    A JSON float compatible with Python's `float`.
    """

    def __round_dump__(self) -> str:
        if self.origin:
            constructed = float(self.origin)
            if (math.isnan(constructed) and math.isnan(self)) or self == constructed:
                return self.origin

        return float.__repr__(self).replace("nan", "NaN").replace("inf", "Infinity")


AnyNumber: TypeAlias = "Integer | Float"

T = TypeVar("T")


class JWrapper(JType, Generic[T]):
    """
    Represents a JSON Literal and wraps the equivalent value in Python.
    """

    value: T
    """
    The Python equivalent value.
    """

    def __init__(self, value: T) -> None:
        self.value = value

    def __eq__(self, obj: object) -> bool:
        return self.value.__eq__(obj.value if isinstance(obj, JWrapper) else obj)

    def __hash__(self) -> int:
        return self.value.__hash__()


Value: TypeAlias = "JObject | Array | JString | JNumber | JWrapper | bool | None"
"""
A type alias matching the JSON Value.
"""

Member = Tuple[Key, Value]
"""
A Key-Value pair in an [Object][kayaku.backend.types.Object]
"""


JSONType_T = TypeVar("JSONType_T", bound=JType)


@overload
def convert(obj: dict) -> JObject:
    ...


@overload
def convert(obj: "list | tuple") -> Array:
    ...


@overload
def convert(obj: str) -> JString:
    ...


@overload
def convert(obj: int) -> Integer:
    ...


@overload
def convert(obj: float) -> Float:
    ...


@overload
def convert(obj: "bool") -> JWrapper[bool]:
    ...


@overload
def convert(obj: None) -> JWrapper[None]:
    ...


@overload
def convert(obj: JSONType_T) -> JSONType_T:
    ...


def convert(obj: Any) -> JType:
    if isinstance(obj, JType):
        return obj
    if isinstance(obj, (list, tuple)):
        o = Array(obj)
    elif (
        isinstance(obj, (bool, date, time, datetime, re.Pattern, enum.Enum))
        or obj == None
    ):
        o = JWrapper(obj)
    elif isinstance(obj, dict):
        o = JObject(obj)
    elif isinstance(obj, str):
        o = JString(obj)
    elif isinstance(obj, int):
        o = Integer(obj)
    elif isinstance(obj, float):
        o = Float(obj)
    else:
        raise TypeError(f"{obj} can't be automatically converted to JSONType!")
    o.__post_init__()
    return o



"""
This module handles whitespaces and comments
"""

from __future__ import annotations

from lark.lark import Lark
from lark.lexer import Token
from lark.visitors import Transformer, v_args

from .types import WSC, BlockStyleComment, LineStyleComment, WhiteSpace


class WSCTransformer(Transformer):
    """
    A [Transformer][lark.visitors.Transformer] handling whitespaces and comments.
    """

    @v_args(inline=True)
    def WS(self, token: Token) -> WhiteSpace:
        return WhiteSpace(token.value)

    def CPP_COMMENT(self, token: Token) -> LineStyleComment:
        return LineStyleComment(token.value[2:])

    def C_COMMENT(self, token: Token) -> BlockStyleComment:
        return BlockStyleComment(token.value[2:-2])

    def wscs(self, wscs: list[WSC]) -> list[WSC]:
        return wscs


transformer = WSCTransformer()


def encode_wsc(wsc: WSC):
    """
    Encode a [WSC][kayaku.backend.types.WSC] into its string representation.

    :param wsc: The Whitespace or Comment to encode.
    """
    if isinstance(wsc, LineStyleComment):
        return f"//{wsc}"
    if isinstance(wsc, BlockStyleComment):
        return f"/*{wsc}*/"
    if isinstance(wsc, WhiteSpace):
        return str(wsc)
    raise NotImplementedError(f"Unknown whitespace or comment type: {wsc!r}")


parser = Lark.open(
    "grammar/wsc.lark",
    rel_to=__file__,
    lexer="basic",
    parser="lalr",
    start="wscs",
    maybe_placeholders=False,
    regex=True,
    transformer=transformer,
)



import inspect
from dataclasses import dataclass, field
from typing import Any, Dict, List

import pytest

from kayaku import backend as json5
from kayaku.doc_parse import store_field_description
from kayaku.format import format_with_model
from kayaku.pretty import Prettifier


def test_format_model():
    @dataclass
    class D:
        a: int = 5

    @dataclass
    class Model:
        c: List[str]
        "Fantasy C"
        e: Any
        "Any E"
        a: int = 4
        """Annotation: A"""
        b: Dict[str, str] = field(default_factory=lambda: {"a": "c"})
        """B
        Annotation: B
        """
        default: D = field(default_factory=D)

    store_field_description(Model, Model.__dataclass_fields__)

    data = json5.loads(
        """
{
    /*
     * Annotation: A
     *
     * @type: int
     */
    a: 3,
    d: 5,
    c: ["123"]
    /*
     * B
     * Annotation: B
     *
     * @type: Dict[str, str]
     */
}
"""
    )

    format_with_model(data, Model)

    assert json5.dumps(Prettifier().prettify(data)) == inspect.cleandoc(
        """\
        {
            /*
             * Annotation: A
             *
             * @type: int
             */
            "a": 3,
            "d": 5,
            /*
             * Fantasy C
             *
             * @type: List[str]
             */
            "c": ["123"],
            /*
             * Any E
             *
             * @type: Any
             */
            "e": null,
            /*
             * B
             * Annotation: B
             *
             * @type: Dict[str, str]
             */
            "b": {"a": "c"},
            /*@type: format.test_format_model.<locals>.D*/
            "default": {"a": 5}
        }
        """
    )

    with pytest.raises(TypeError):
        format_with_model({}, Model)
